# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import collections
import inspect
import typing
from typing import Optional, Sequence, Union

import onnx

from onnxscript import onnx_types

# TypeAnnotationValue represents the (value of) valid type-annotations recognized
# by ONNX Script. TODO: Flesh out a formal definition. Currently, it supports
# - float, int, str (primitive attribute types)
# - Sequence[float], Sequence[int], Sequence[str] (attribute types)
# - Tensor types
# - Sequence[Tensor] types
# - Union of above 2
# - TypeVars with above bounds
# - Above types with annotation attached
TypeAnnotationValue = typing.Any


_PRIMITIVE_ATTR_TYPE = frozenset((float, int, str, bool))

_LIST_CONSTRUCTORS = frozenset([list, typing.List, typing.Sequence, collections.abc.Sequence])

# Map from ONNX AttributeProto type to its representation (in ONNX Script).
_ATTRTYPE_TO_REPR = {
    onnx.AttributeProto.FLOAT: "float",
    onnx.AttributeProto.INT: "int",
    onnx.AttributeProto.STRING: "str",
    onnx.AttributeProto.FLOATS: "Sequence[float]",
    onnx.AttributeProto.INTS: "Sequence[int]",
    onnx.AttributeProto.STRINGS: "Sequence[str]",
}


def onnx_attr_type_to_onnxscript_repr(attr_type: onnx.AttributeProto.AttributeType) -> str:
    if attr_type not in _ATTRTYPE_TO_REPR:
        supported = ", ".join(
            f"'{onnx.AttributeProto.AttributeType.Name(v)}'" for v in _ATTRTYPE_TO_REPR
        )
        raise ValueError(f"Unsupported attribute type {attr_type}: only {supported} allowed.")
    return _ATTRTYPE_TO_REPR[attr_type]


# A sorted list of all type strings used in an OpSchema
ALL_TENSOR_TYPE_STRINGS = tuple(
    sorted(
        tensor_type.to_string()
        for tensor_type in onnx_types.tensor_type_registry.values()
        # Skip FLOAT4E2M1 for versions older than 1.18, and FLOAT8E8M0 for versions older than 1.19
        # TODO(after onnx requirement bump): Remove this check
        if (
            not (
                not hasattr(onnx.TensorProto, "FLOAT4E2M1")
                and tensor_type == onnx_types.FLOAT4E2M1
            )
            and not (
                not hasattr(onnx.TensorProto, "FLOAT8E8M0")
                and tensor_type == onnx_types.FLOAT8E8M0
            )
        )
    )
)


def _remove_annotation(typeinfo: TypeAnnotationValue) -> TypeAnnotationValue:
    """Remove Annotated wrapper if present, otherwise return typeinfo as is."""
    if hasattr(typing, "Annotated"):
        # Present in Python 3.9+
        if typing.get_origin(typeinfo) is typing.Annotated:
            return typing.get_args(typeinfo)[0]
    return typeinfo


def _is_primitive_attr_type(typeinfo: TypeAnnotationValue) -> bool:
    return typeinfo in _PRIMITIVE_ATTR_TYPE


def base_type_is_bool(pytype: TypeAnnotationValue) -> bool:
    """Returns True if base type of pytype is bool, False otherwise."""
    pytype = _remove_annotation(pytype)
    if pytype in _PRIMITIVE_ATTR_TYPE:
        return pytype is bool
    type_constructor = typing.get_origin(pytype)
    if type_constructor in _LIST_CONSTRUCTORS:
        element_type = typing.get_args(pytype)[0]
        return element_type is bool
    # Remove Optional wrapper if present:
    if type_constructor is Optional or type_constructor is Union:
        # In Python < 3.10, Optional[X] is represented as Union[X, type(None)]
        # so we filter out type(None) if present
        args = [x for x in typing.get_args(pytype) if x is not type(None)]
        if len(args) == 1:
            return base_type_is_bool(args[0])

    return False


def _is_tensor_type(typeinfo: TypeAnnotationValue) -> bool:
    if isinstance(typeinfo, onnx_types.TensorType):
        return True
    if inspect.isclass(typeinfo) and issubclass(typeinfo, onnx_types.TensorType):
        return True
    return False


def is_value_type(typeinfo: TypeAnnotationValue) -> bool:
    """Returns True if typeinfo represents a value type, False if it is an attribute type.
    Raises ValueError if typeinfo is not a supported type annotation.
    """
    typeinfo = _remove_annotation(typeinfo)
    if _is_tensor_type(typeinfo):
        return True
    if _is_primitive_attr_type(typeinfo):
        return False
    type_constructor = typing.get_origin(typeinfo)
    # Handle List-like type-constructor
    # Eg. List[INT32] is a value type, while List[int] is an attribute type
    if type_constructor in _LIST_CONSTRUCTORS:
        elt_type = typing.get_args(typeinfo)[0]
        return is_value_type(elt_type)
    # Handle Union and Optional type-constructors
    if type_constructor is typing.Union:
        # Filter out None, since typing.Optional[X] evaluates to Union[X, None]
        args = [x for x in typing.get_args(typeinfo) if x is not type(None)]
        args_value_check = [is_value_type(x) for x in args]
        if all(args_value_check):
            # Handles cases like Optional[INT32] as well as Union[FLOAT16, FLOAT, DOUBLE]
            return True
        elif (len(args) == 1) and args_value_check[0] is False:
            # Handle the case of optional attribute: eg. Optional[int]
            # Note that we do not allow Union[int, float] for attributes.
            return False
        else:
            raise ValueError(f"Unsupported type annotation '{typeinfo}'")
    # Handle TypeVars:
    if isinstance(typeinfo, typing.TypeVar):
        if hasattr(typeinfo, "__bound__"):
            bound = typeinfo.__bound__
            return is_value_type(bound)
    raise ValueError(f"Unsupported type annotation {typeinfo}")


def is_attr_type(pytype: TypeAnnotationValue):
    return is_value_type(pytype) is False


def is_valid_type(typeinfo: TypeAnnotationValue):
    try:
        return is_value_type(typeinfo) in {True, False}
    except ValueError:
        return False


def is_optional(pytype) -> bool:
    """Returns whether a pytype is an Optional."""
    if typing.get_origin(pytype) is Union and type(None) in typing.get_args(pytype):
        # Python < 3.10
        return True
    if typing.get_origin(pytype) is Optional:
        # Python >= 3.10
        return True
    return False


def get_return_types(typeinfo: type | Sequence[type]) -> Sequence[type]:
    """Converts return-type annotation into a sequence of types.

    The return type annotation can be either a single type (for a single output)
    or a Tuple type (for multiple outputs). This function normalizes the
    representation so that it is always a sequence of types, even for a single
    output.
    """
    if isinstance(typeinfo, typing.Sequence):
        return typeinfo
    if typing.get_origin(typeinfo) is tuple:
        return typing.get_args(typeinfo)
    return (typeinfo,)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""The Pattern IR: used to describe (source) patterns of rewrite rules."""

from __future__ import annotations

import abc
import contextlib
import inspect
import itertools
from collections.abc import Mapping
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

import onnxscript.rewriter._basics as _basics
from onnxscript import ir

T = TypeVar("T")


class Pattern(Protocol[T]):  # type: ignore[misc]
    """This is essentially a Predicate[T], that is, a Callable[[T], bool] bound to the name "matches"."""

    def matches(self, item: T) -> bool: ...


class StringPattern(abc.ABC, Pattern[str]):
    """Abstract base class for string patterns."""

    @abc.abstractmethod
    def matches(self, item: str) -> bool:
        pass

    @abc.abstractmethod
    def __str__(self) -> str:
        pass


class StringConstantPattern(StringPattern):
    """Matches strings with given value."""

    def __init__(self, value: str):
        self._value = value

    def matches(self, item: str) -> bool:
        return item == self._value

    def __str__(self) -> str:
        return self._value

    def value(self) -> str:
        return self._value


class PrefixPattern(StringPattern):
    """Matches strings with a given prefix."""

    def __init__(self, value: str) -> None:
        self._value = value

    def matches(self, value: str) -> bool:
        return value.startswith(self._value)

    def __str__(self) -> str:
        return f"{self._value}*"


class AttrPattern(Pattern[ir.Attr]):
    """Base class for an attribute pattern. Matches any attribute value by default."""

    def __init__(self, name: str | None, *, can_match_none: bool = False):
        self._name = name
        self._can_match_none = can_match_none

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def can_match_none(self) -> bool:
        """Indicates whether this pattern can match a None attribute."""
        return self._can_match_none

    def matches(self, attr: ir.Attr) -> bool:
        return True

    def __str__(self) -> str:
        return self._name if self._name is not None else "anonymous:" + str(id(self))


class AttrVar(AttrPattern):
    """Represents a pattern variable used to match against attribute values."""

    def __init__(self, name: str | None, *, can_match_none: bool = False):
        super().__init__(name, can_match_none=can_match_none)


# TODO: Support tensors. Align with usage elsewhere.
SupportedAttrTypes = Union[
    int,
    float,
    str,
    Sequence[int],
    Sequence[float],
    Sequence[str],
]


class AttrConstantPattern(AttrPattern):
    """Matches attributes with given value.

    Uses standard equality for matching. For list-valued attributes, the order of elements matters.
    If order is immaterial, we need to define a separate pattern for that.
    """

    def __init__(self, value: SupportedAttrTypes):
        super().__init__(None)
        self._value = value

    def matches(self, attr: ir.Attr) -> bool:
        if attr.type in {
            ir.AttributeType.INTS,
            ir.AttributeType.FLOATS,
            ir.AttributeType.STRINGS,
        }:
            # Since the type of attr.value is Sequence, we need to convert to the same type for comparison.
            return tuple(attr.value) == tuple(self._value)
        return attr.value == self._value

    def __str__(self) -> str:
        return str(self._value)


def _to_attr_pattern(value: AttrPattern | ValuePattern | SupportedAttrTypes) -> AttrPattern:
    """Represents promotion of values allowed as keyword-arguments in a pattern-builder call to an AttrPattern."""
    if isinstance(value, AttrPattern):
        return value
    if isinstance(value, Var):
        # This is a hack. Currently, when we create pattern-variables, we create them as Var,
        # and change them to AttrPattern if/when used in an attribute context. We could use type
        # annotations to distinguish between ValuePattern and AttrPattern, but forces users to
        # use these type annotations.
        # TODO: check for misuse at rule-creation time. (Currently will be caught by matcher at match-time.)
        if value.check_method is not None:
            raise ValueError(
                "Pattern variables used in attributes must not have check_method set."
            )
        return AttrVar(value.name, can_match_none=value.can_match_none)
    if isinstance(value, (int, float, str)):
        return AttrConstantPattern(value)
    if isinstance(value, Sequence):
        if all(isinstance(i, (int, float)) for i in value):
            return AttrConstantPattern(value)
        if all(isinstance(i, str) for i in value):
            return AttrConstantPattern(value)
        raise ValueError("Only lists of int/float/str can be used as an AttrPattern")
    raise TypeError(f"Cannot convert {type(value)} to AttrPattern")


class OpsetPatternBuilder:
    """Represents an opset pattern and a pattern builder.

    (i) It is used to create a NodePattern (via OpPatternBuilder).
    Example usage:
    ::

        z = op.Matmul(x, y)

    Here, `op` is an instance of OpsetPatternBuilder and `op.Matmul` is an instance
    of OpPatternBuilder, and  `op.Matmul(x, y)` is an instance of NodePattern.

    (ii) It contains a domain pattern matched against the actual opset domain used in the
    input model.
    """

    def __init__(self, domain: StringPattern | str, record: bool = False) -> None:
        if isinstance(domain, str):
            domain = StringConstantPattern(domain)
        self._domain_pattern = domain
        if record:
            self._nodes: list[NodePattern] | None = []
        else:
            self._nodes = None

    def domain_pattern(self) -> StringPattern:
        return self._domain_pattern

    def __getattr__(self, op_name: str) -> OpPatternBuilder:
        return OpPatternBuilder(self, op_name)

    def submodule(self, name: str) -> OpPatternBuilder:
        """This method is used to match against submodule ops with prefix."""
        return OpPatternBuilder(self, PrefixPattern(name))

    def __str__(self) -> str:
        return str(self._domain_pattern)

    def add_node(self, node: NodePattern) -> None:
        if self._nodes is not None:
            self._nodes.append(node)

    def nodes(self) -> Sequence[NodePattern]:
        if self._nodes is None:
            raise ValueError("Nodes were not recorded.")
        return self._nodes


onnxop = OpsetPatternBuilder("")

torch_module_op = OpsetPatternBuilder(PrefixPattern("pkg.torch"))


class OpPatternBuilder:
    """A utility class to build a NodePattern.

    It is used primarily to create a NodePattern.
    Example usage:
    ::

        z = op.Matmul(x, y)

    Here, `op` is an instance of OpsetPatternBuilder and `op.Matmul` is an instance
    of OpPatternBuilder, and  `op.Matmul(x, y)` is an instance of NodePattern.

    """

    def __init__(
        self,
        pattern_builder: OpsetPatternBuilder,
        op_name: str | Pattern[str],
    ) -> None:
        self.pattern_builder = pattern_builder
        self.op_name = op_name

    def __call__(
        self,
        *args,
        _domain: str | None = None,
        _version: int | None = None,
        _outputs: int | list[str | None] = 1,
        _allow_other_attributes: bool | None = None,
        _allow_other_inputs: bool | None = None,
        _check: Callable | None = None,
        **kwargs,
    ):
        if _version is not None:
            raise ValueError(
                "The pattern builder does not support '_version' keyword argument. "
                "Version restrictions should be handled by rewrite rules."
            )
        if _domain is None:
            opset_pattern = self.pattern_builder.domain_pattern()
        elif isinstance(_domain, str):
            opset_pattern = StringConstantPattern(_domain)
        else:
            # TODO(rama): allow OpsetPatternBuilder as _domain.
            raise TypeError("_domain must be a string.")

        if isinstance(_outputs, int):
            _outputs = [None for _ in range(_outputs)]
        elif not isinstance(_outputs, Sequence) or not all(
            isinstance(x, (str, type(None))) for x in _outputs
        ):
            raise ValueError("_outputs must be an int or a list[str|None].")
        inputs = [_to_value_pattern(x) for x in args]
        attributes = {name: _to_attr_pattern(value) for (name, value) in kwargs.items()}
        node_pattern = NodePattern(
            opset_pattern,
            self.op_name,
            inputs,
            attributes,
            _outputs,
            allow_other_attributes=_allow_other_attributes,
            allow_other_inputs=_allow_other_inputs,
            check=_check,
        )
        self.pattern_builder.add_node(node_pattern)
        output_values = node_pattern.outputs
        # Unpack outputs if there is only one output, the common case.
        if len(output_values) == 1:
            return output_values[0]
        else:
            return output_values


def _to_value_pattern(
    x: ValuePattern | int | float | Callable | None,
) -> ValuePattern | None:
    """Promotes an input-value used to construct a NodePattern to a ValuePattern.

    Example usage:
    ::
        x = op.MatMul(a, b)
        z = op.Add(x, 0)

    In this example, `a, `b`, and `x` are ValuePatterns used to construct a NodePattern.
    `0` is a constant (int) value, and is automatically promoted to a ValuePattern.

    Note that this is a shorthand for creating a Constant pattern. The user can more
    explicitly write this as:
    ::
        z = op.Add(x, op.Constant(0))

    If a callable is provided, it will be converted to a ValuePattern with the callable as the check attribute.
    """
    if x is None or isinstance(x, ValuePattern):
        return x
    if isinstance(x, (int, float)):
        return Constant(x)
    if isinstance(x, Sequence):
        if all(isinstance(i, (int, float)) for i in x):
            return Constant(x)
        raise ValueError("Only lists of int/float can be used as a ValuePattern")
    if callable(x):
        return ValuePattern(None, check=x)

    raise TypeError(f"Cannot convert {type(x)} to ValuePattern")


_pattern_builder: OpsetPatternBuilder = onnxop


@contextlib.contextmanager
def pattern_builder(builder: OpsetPatternBuilder):
    global _pattern_builder
    prev_builder = _pattern_builder
    _pattern_builder = builder
    yield
    _pattern_builder = prev_builder


class ValuePattern:
    """Base class for all patterns that match against IR values.

    This is used primarily to provide operator overloadings for arithmetic
    operations, so that we can write patterns like `x + 1` and `1 + x`.
    """

    def __init__(
        self, name: str | None, *, check: Callable | None = None, can_match_none: bool = False
    ) -> None:
        self._name = name
        self._check = check
        self._can_match_none = can_match_none
        # Note: uses will be computed only when the full graph-pattern is constructed.
        self._uses: list[tuple[NodePattern, int]] = []

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> ValuePattern:
        del node_map
        return ValuePattern(self._name, check=self._check)

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def check_method(self) -> Callable | None:
        return self._check

    @property
    def can_match_none(self) -> bool:
        """Indicates whether this variable can match a None input."""
        return self._can_match_none

    def producer(self) -> NodePattern | None:
        return None

    def uses(self) -> Sequence[tuple[NodePattern, int]]:
        return self._uses

    def append_use(self, node: NodePattern, index: int):
        self._uses.append((node, index))

    def __repr__(self) -> str:
        return f"ValuePattern({self._name!r})"

    def __add__(self, other):
        return _pattern_builder.Add(self, other)

    def __radd__(self, other):
        return _pattern_builder.Add(other, self)

    def __sub__(self, other):
        return _pattern_builder.Sub(self, other)

    def __rsub__(self, other):
        return _pattern_builder.Sub(other, self)

    def __mul__(self, other):
        return _pattern_builder.Mul(self, other)

    def __rmul__(self, other):
        return _pattern_builder.Mul(other, self)

    def __truediv__(self, other):
        return _pattern_builder.Div(self, other)

    def __rtruediv__(self, other):
        return _pattern_builder.Div(other, self)

    def __pow__(self, other):
        return _pattern_builder.Pow(self, other)

    def __str__(self) -> str:
        return self._name if self._name is not None else "anonymous:" + str(id(self))


class NodePattern:
    """Represents a pattern that matches against a Node.

    This differs from a NodeOutputPattern in that it matches against a node (which
    may produce 1 or more outputs), whereas a NodeOutputPattern matches against
    a specific output of a node.

    Args:
        domain: pattern to match against the domain of the node.
        op: pattern or string constant to match against the op_type of the node.
        inputs: sequence of ValuePatterns (or constants) to match against the inputs of the node.
        attributes: dictionary of attribute patterns to match against the attributes of the node.
        outputs: specifies pattern-variable-name for outputs (or None)
        allow_other_attributes: specifies whether other attributes (not mentioned in `attributes`)
          are allowed in the node.
    """

    def __init__(
        self,
        domain: StringPattern,
        op: str | Pattern[str],
        inputs: Sequence[int | float | ValuePattern | None],
        attributes: dict[str, AttrPattern],
        outputs: Sequence[str | None],
        *,
        allow_other_attributes: bool | None,
        allow_other_inputs: bool | None,
        check: Callable | None = None,
    ):
        if allow_other_attributes is None:
            # Default behavior: allow other unmatched attributes in the node.
            allow_other_attributes = True
        if allow_other_inputs is None:
            # TODO(rama): Should we default to True? For now, we preserve the current behavior.
            allow_other_inputs = False
        self.domain = domain
        self.op = StringConstantPattern(op) if isinstance(op, str) else op
        self.inputs = [_to_value_pattern(x) for x in inputs]
        self.attributes = attributes
        self.allow_other_attributes = allow_other_attributes
        self.allow_other_inputs = allow_other_inputs
        self._check = check
        # In the common case, domain and op are constants, which can be used to optimize matching.
        if isinstance(op, str) and isinstance(domain, StringConstantPattern):
            # TODO(rama): support overloaded operators.
            overload = ""
            self._op_identifier: ir.OperatorIdentifier | None = (
                domain.value(),
                op,
                overload,
            )
        else:
            self._op_identifier = None
        self.outputs = [NodeOutputPattern(self, i, name) for i, name in enumerate(outputs)]

        # Update uses for inputs.
        for index, value in enumerate(self.inputs):
            if value is not None:
                value.append_use(self, index)

    def __str__(self) -> str:
        inputs = ", ".join(str(v) for v in self.inputs)
        outputs = ", ".join(str(v) for v in self.outputs)
        attributes = ", ".join(f"{k}={v}" for k, v in self.attributes.items())
        op = str(self.op)
        domain = str(self.domain)
        qualified_op = f"{domain}.{op}" if domain else op
        inputs_and_attributes = f"{inputs}, {attributes}" if attributes else inputs
        return f"{outputs} = {qualified_op} ({inputs_and_attributes})"

    def op_identifier(self) -> ir.OperatorIdentifier | None:
        return self._op_identifier

    @property
    def op_type(self) -> str:
        return str(self.op)

    @property
    def check_method(self) -> Callable | None:
        return self._check

    def matches(self, node: ir.Node, match: _basics.MatchResult) -> _basics.MatchResult:
        """Matches the pattern represented by self against a node.

        This is purely a local node-level match, and does not consider the subgraph rooted at the node.
        We check the domain, op_type, and attributes of the node, but not the inputs.
        """
        # TODO(rama): Ensure we handle "" and "onnx.ai" correctly.
        if not self.op.matches(node.op_type):
            return match.fail(
                f"OpType mismatch: expected {self.op}, got {node.op_type}.", node
            )
        if not self.domain.matches(node.domain):
            return match.fail(
                f"Domain mismatch: expected {self.domain}, got {node.domain}.", node
            )

        for name, attr_pattern in self.attributes.items():
            attr_value = node.attributes.get(name)
            if attr_value is None:
                if not attr_pattern.can_match_none:
                    return match.fail(f"Attribute {name} not found in node.", node)
            elif not attr_pattern.matches(attr_value):
                return match.fail(
                    f"Attribute {name} mismatch: expected {attr_pattern}, got {attr_value}.",
                    node,
                )
            if attr_pattern.name is not None:
                if not match.bind(attr_pattern.name, attr_value):
                    return match

        if not self.allow_other_attributes:
            for name in node.attributes:
                # TODO: Support matching default nodes for attributes.
                if name not in self.attributes:
                    return match.fail(f"Attribute {name} not expected in node.", node)

        return match

    def clone(self, node_map: dict[NodePattern, NodePattern], swap: bool) -> NodePattern:
        inputs = [(v.clone(node_map) if v is not None else None) for v in self.inputs]
        if swap:
            assert len(inputs) == 2, (
                "Internal error: commutative swap applies only to binary ops."
            )
            inputs = [inputs[1], inputs[0]]
        outputs = [value.name for value in self.outputs]
        copied = NodePattern(
            self.domain,
            self.op,
            inputs,
            self.attributes,
            outputs,
            allow_other_attributes=self.allow_other_attributes,
            allow_other_inputs=self.allow_other_inputs,
            check=self._check,
        )
        node_map[self] = copied
        return copied


class NodeOutputPattern(ValuePattern):
    """Represents a pattern that matches against a specific output of a Node.

    This is the primary pattern used to match against computed values, that
    is values computed using a specific op.
    """

    def __init__(
        self, producer: NodePattern, output_index: int, name: str | None = None
    ) -> None:
        super().__init__(name)
        self._producer = producer
        self._output_index = output_index

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> NodeOutputPattern:
        return node_map[self._producer].outputs[self._output_index]
        # return NodeOutputPattern(node_map[self._producer], self._output_index, self._name)

    @property
    def output_index(self) -> int:
        return self._output_index

    def producer(self) -> NodePattern:
        return self._producer


class Var(ValuePattern):
    """Represents a pattern-variable."""

    def __init__(
        self, name: str | None, *, check: Callable | None = None, can_match_none: bool = False
    ) -> None:
        super().__init__(name, check=check, can_match_none=can_match_none)

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> Var:
        """Clones the pattern-variable, preserving its name and check method."""
        return Var(self.name, check=self.check_method, can_match_none=self.can_match_none)


class AnyValue(ValuePattern):
    """Represents a pattern that matches against any value."""

    def __init__(self) -> None:
        super().__init__(None)

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> AnyValue:
        # A single instance of AnyValue suffices.
        return self


ANY_VALUE = AnyValue()


class Constant(ValuePattern):
    """Represents a pattern that matches against a scalar constant value."""

    def __init__(
        self,
        value: int | float | Sequence[int] | Sequence[float],
        rel_tol: float = 1e-5,
        abs_tol: float = 1e-8,
    ) -> None:
        super().__init__(None)
        self._value = list(value) if isinstance(value, Sequence) else value
        self._rel_tol = rel_tol
        self._abs_tol = abs_tol

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> Constant:
        del node_map
        return Constant(self._value, self._rel_tol, self._abs_tol)

    @property
    def value(self) -> int | float | list[int] | list[float]:
        return self._value

    def __str__(self) -> str:
        return str(self._value)


class OpIdDispatchOr(ValuePattern):
    """Represents a (restricted) form of value pattern disjunction that enables deterministic matching."""

    def __init__(
        self,
        op_to_pattern: Mapping[ir.OperatorIdentifier, tuple[Any, ValuePattern]],
        name: str | None = None,
        tag_var: str | None = None,
    ) -> None:
        """
        Initialize an OpIdDispatchOr pattern.

        Args:
            op_to_pattern: A dictionary mapping operator identifiers to tuples of tag values and patterns.
                The keys are operator identifiers, and the values are tuples containing a tag value
                and a pattern to match against.
            name: An optional variable name for the pattern. Defaults to None. If present,
                this name will be bound to the value matched by the pattern.
            tag_var: An optional variable name for the tag. Defaults to None. If present,
                it will be bound to a value indicating which alternative was matched.
        """
        super().__init__(name)
        self._op_to_pattern = op_to_pattern
        self._tag_var = tag_var

    @property
    def tag_var(self) -> str | None:
        """Returns the tag variable associated with the OrValue pattern."""
        return self._tag_var

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> OpIdDispatchOr:
        return OpIdDispatchOr(
            {k: (v[0], v[1].clone(node_map)) for k, v in self._op_to_pattern.items()},
            self.name,
            self._tag_var,
        )

    def get_pattern(self, value: ir.Value) -> tuple[Any, ValuePattern] | None:
        """Returns the pattern that should be tried for the given value."""
        producer = value.producer()
        if producer is not None:
            id = producer.op_identifier()
            if id is not None and id in self._op_to_pattern:
                return self._op_to_pattern[id]
        return None


class BacktrackingOr(ValuePattern):
    """Represents an unrestricted form of OR pattern implemented using backtracking."""

    def __init__(
        self,
        values: Sequence[ValuePattern],
        name: str | None = None,
        tag_var: str | None = None,
        tag_values: Sequence[Any] | None = None,
    ) -> None:
        """
        Initialize a BacktrackingOr pattern.

        Args:
            values: A sequence of value patterns to match against.
            name: An optional variable name for the pattern. Defaults to None. If present,
                this name will be bound to the value matched by the pattern.
            tag_var: An optional variable name for the tag. Defaults to None. If present,
                it will be bound to a value (from tag_values) indicating which alternative was matched.
            tag_values: An optional sequence of values to bind to the tag_var. Defaults to None.
                If present, the length of tag_values must match the number of alternatives in values.
                In a successful match, tag-var will be bound to the i-th value in tag_values if the i-th
                alternative pattern matched. If omitted, the default value of (0, 1, 2, ...) will be used.
        """
        super().__init__(name)
        if tag_values is not None:
            if tag_var is None:
                raise ValueError("tag_var must be specified if tag_values is provided.")
            if len(tag_values) != len(values):
                raise ValueError(
                    "tag_values must have the same length as the number of alternatives."
                )
        else:
            tag_values = tuple(range(len(values)))
        self._tag_var = tag_var
        self._tag_values = tag_values
        self._values = values

    @property
    def tag_var(self) -> str | None:
        """Returns the tag variable associated with the OrValue pattern."""
        return self._tag_var

    def clone(self, node_map: dict[NodePattern, NodePattern]) -> BacktrackingOr:
        return BacktrackingOr(
            [v.clone(node_map) for v in self._values],
            self.name,
            self._tag_var,
            self._tag_values,
        )


def OrValue(
    values: Sequence[ValuePattern],
    name: str | None = None,
    tag_var: str | None = None,
    tag_values: Sequence[Any] | None = None,
) -> ValuePattern:
    """
    Creates an OR pattern.

    Args:
        values: A sequence of value patterns to match against.
        name: An optional variable name for the pattern. Defaults to None. If present,
            this name will be bound to the value matched by the pattern.
        tag_var: An optional variable name for the tag. Defaults to None. If present,
            it will be bound to a value (from tag_values) indicating which alternative was matched.
        tag_values: An optional sequence of values to bind to the tag_var. Defaults to None.
            If present, the length of tag_values must match the number of alternatives in values.
            In a successful match, tag-var will be bound to the i-th value in tag_values if the i-th
            alternative pattern matched. If omitted, the default value of (0, 1, 2, ...) will be used.
    """
    if tag_values is not None:
        if tag_var is None:
            raise ValueError("tag_var must be specified if tag_values is provided.")
        if len(tag_values) != len(values):
            raise ValueError(
                "tag_values must have the same length as the number of alternatives."
            )
    else:
        tag_values = tuple(range(len(values)))

    def make_op_id_or_pattern() -> OpIdDispatchOr | None:
        mapping: dict[ir.OperatorIdentifier, tuple[Any, NodeOutputPattern]] = {}
        for i, alternative in enumerate(values):
            if not isinstance(alternative, NodeOutputPattern):
                return None
            producer = alternative.producer()
            id = producer.op_identifier()
            if id is None or id in mapping:
                return None
            mapping[id] = (tag_values[i], alternative)
        return OpIdDispatchOr(mapping, name, tag_var)

    optimized_pattern = make_op_id_or_pattern()
    return optimized_pattern or BacktrackingOr(
        values, name, tag_var, tag_values if tag_var else None
    )


def _nodes_in_pattern(outputs: Sequence[ValuePattern]) -> list[NodePattern]:
    """Returns all nodes used in a pattern, given the outputs of the pattern."""
    node_patterns: list[NodePattern] = []

    def visit(value_patterns: Sequence[ValuePattern | None]) -> None:
        for value_pattern in value_patterns:
            if isinstance(value_pattern, NodeOutputPattern):
                node_pattern = value_pattern.producer()
                if node_pattern not in node_patterns:
                    node_patterns.append(node_pattern)
                    visit(node_pattern.inputs)

    visit(outputs)
    node_patterns.reverse()
    return node_patterns


def _add_backward_slice(
    node: NodePattern,
    backward_slice: set[NodePattern],
    backward_slice_values: set[ValuePattern],
) -> None:
    """Adds all nodes in the backward slice of given node to the set `backward_slice`.

    The backward slice of a node is the set of all nodes that are reachable from the node
    in a backward traversal from the given node.
    """
    if node in backward_slice:
        return
    backward_slice.add(node)
    for value_pattern in node.inputs:
        if isinstance(value_pattern, NodeOutputPattern):
            _add_backward_slice(
                value_pattern.producer(), backward_slice, backward_slice_values
            )
        elif isinstance(value_pattern, (OpIdDispatchOr, BacktrackingOr)):
            backward_slice_values.add(value_pattern)


class GraphPattern:
    """Represents a pattern that can be matched against a subgraph."""

    def __init__(
        self,
        inputs: Sequence[ValuePattern],
        outputs: Sequence[ValuePattern],
        nodes: Sequence[NodePattern],
    ) -> None:
        self._inputs = inputs
        self._outputs = outputs
        if len(outputs) == 0:
            raise ValueError("GraphPattern must have at least one output")
        self._nodes = nodes  # _nodes_in_pattern(outputs)

        # Determine the output nodes of the pattern. These are a minimal set of nodes
        # whose backward-slices cover the entire pattern.
        # Use a dict as an ordered set to preserve deterministic insertion order
        # from the outputs sequence. Using a plain set would cause non-deterministic
        # ordering due to Python's hash randomization, leading to non-deterministic
        # pattern matching behavior.
        output_nodes: dict[NodePattern, None] = {}
        covered: set[NodePattern] = set()
        choice_values_returned: set[ValuePattern] = set()
        covered_choice_values: set[ValuePattern] = set()
        for value_pattern in outputs:
            if not isinstance(value_pattern, ValuePattern):
                raise TypeError(
                    f"Invalid type {type(value_pattern)} for graph pattern output."
                )
            if isinstance(value_pattern, NodeOutputPattern):
                candidate = value_pattern.producer()
                if candidate not in covered:
                    output_nodes[candidate] = None
                    _add_backward_slice(candidate, covered, covered_choice_values)
            elif isinstance(value_pattern, (OpIdDispatchOr, BacktrackingOr)):
                choice_values_returned.add(value_pattern)

        # check if all choice_values_returned are contained in covered_choice_values:
        # We don't yet support the use of a choice-value as a "root" of the search.
        # This is a limitation of the current implementation, and will be fixed in the future.
        if not (choice_values_returned <= covered_choice_values):
            raise NotImplementedError("Returning uncovered choice-values is not supported.")

        self.output_nodes: list[NodePattern] = list(output_nodes)

    @property
    def output_node(self) -> NodePattern:
        if len(self.output_nodes) != 1:
            raise ValueError("GraphPattern does not have unique output node.")
        return self.output_nodes[0]

    def node(self, index: int) -> NodePattern:
        return self._nodes[index]

    def num_nodes(self) -> int:
        return len(self._nodes)

    def __len__(self) -> int:
        return self.num_nodes()

    @property
    def inputs(self) -> Sequence[ValuePattern]:
        return self._inputs

    @property
    def outputs(self) -> Sequence[ValuePattern]:
        return self._outputs

    def __iter__(self) -> Iterator[NodePattern]:
        return iter(self._nodes)

    def __reversed__(self) -> Iterator[NodePattern]:
        return reversed(self._nodes)

    @property
    def has_single_output_node(self) -> bool:
        return len(self.output_nodes) == 1

    @property
    def num_outputs(self) -> int:
        return len(self._outputs)

    def commute(self) -> Sequence[GraphPattern]:
        # List all commutative elementwise (binary) operators for which we
        # consider swapping the inputs
        COMMUTATIVE_OPS = {
            ("", "Add", ""),
            ("", "Mul", ""),
            ("", "And", ""),
            ("", "Or", ""),
            ("", "Xor", ""),
            ("", "BitwiseAnd", ""),
            ("", "BitwiseOr", ""),
            ("", "BitwiseXor", ""),
            ("", "Equal", ""),
            ("", "Max", ""),
            ("", "Mean", ""),
            ("", "Min", ""),
            ("", "Sum", ""),
        }

        def commute_node(node: NodePattern) -> Iterable[bool]:
            if node.op_identifier() in COMMUTATIVE_OPS:
                # Try with and without swapping inputs.
                return [False, True]
            # No swapping of inputs
            return [False]

        iteration_space = [commute_node(node) for node in self._nodes]

        def copy_graph(swap_list: Iterable[bool]) -> GraphPattern:
            if not any(swap_list):
                # No need to swap inputs of any node
                return self
            # Create a copy of the graph, with swapped inputs for the nodes that need it.
            node_map: dict[NodePattern, NodePattern] = {}
            new_inputs = [v.clone(node_map) for v in self._inputs]
            new_nodes = [
                node.clone(node_map, swap) for node, swap in zip(self._nodes, swap_list)
            ]
            new_outputs = [v.clone(node_map) for v in self._outputs]
            return GraphPattern(new_inputs, new_outputs, new_nodes)

        return [copy_graph(swap_list) for swap_list in itertools.product(*iteration_space)]

    def __str__(self) -> str:
        inputs = ", ".join(str(v) for v in self._inputs)
        outputs = ", ".join(str(v) for v in self._outputs)
        nodes = "\n   ".join(str(n) for n in self._nodes)
        return f"pattern ({inputs}) {{\n   {nodes}\n   return {outputs}\n}}"


def _to_graph_pattern(pattern_constructor: Callable) -> GraphPattern:
    """Convert a pattern-construction function to a GraphPattern.

    A pattern-construction function will return values as below:
    ::
        def pattern(op, x: Var, shape1: Var, shape2: Var):
            ...
            return outputs

    We create a pattern graph by creating pattern-variables for each parameter of the function,
    and calling the function. The returned values are normalized to a list of ValuePatterns,
    which represent the outputs of the pattern graph.

    Args:
        pattern_constructor: Callable

    Returns:
        GraphPattern: A representation of the pattern that can be matched against a subgraph.
    """
    _pattern_vars = inspect.signature(pattern_constructor).parameters
    pattern_inputs = [Var(v) for v in _pattern_vars][1:]  # Skip the first parameter
    builder = OpsetPatternBuilder("", record=True)
    with pattern_builder(builder):
        pattern_outputs = pattern_constructor(builder, *pattern_inputs)
    # TODO(rama): classify inputs as value/attribute vars
    # Returned value could be a single ValuePattern or a list of ValuePatterns.
    # Normalize representation to a list of ValuePatterns.
    if isinstance(pattern_outputs, ValuePattern):
        pattern_outputs = [pattern_outputs]
    return GraphPattern(pattern_inputs, pattern_outputs, builder.nodes())



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# mypy: disable-error-code=misc

import collections
import unittest

import parameterized

from onnxscript import ir
from onnxscript._internal import param_manipulation

TEST_INPUT = "TEST_INPUT"


class TestSeparateInputAttributesFromArguments(unittest.TestCase):
    """Unit tests for `param_manipulation.separate_input_attributes_from_arguments`."""

    @parameterized.parameterized.expand(
        [
            (
                "all_positional",
                (TEST_INPUT, 42, 0.0),
                {},
                0.0,
            ),
            (
                "positional_with_default",
                (TEST_INPUT, 42),
                {},
                100.0,
            ),
            (
                "positional_with_default_and_kwargs",
                (TEST_INPUT,),
                {"b": 42},
                100.0,
            ),
            (
                "positional_with_kwargs",
                (TEST_INPUT, 42),
                {"c": 0.0},
                0.0,
            ),
            (
                "positional_input_with_kwargs_attribute",
                (TEST_INPUT,),
                {"b": 42, "c": 0.0},
                0.0,
            ),
            (
                "all_kwargs",
                (),
                {"a": TEST_INPUT, "b": 42, "c": 0.0},
                0.0,
            ),
            (
                "all_kwargs_with_default",
                (),
                {"a": TEST_INPUT, "b": 42},
                100.0,
            ),
        ]
    )
    def test_it_is_correct_on(self, _, args, kwargs, expected_c):
        # Create OpSignature with one input and two attributes
        type_constraint = ir.schemas.TypeConstraintParam.any_tensor("T")
        op_signature = ir.schemas.OpSignature(
            domain="",
            name="TestOp",
            overload="",
            params=[
                ir.schemas.Parameter(
                    name="a", type_constraint=type_constraint, required=True, variadic=False
                ),
                ir.schemas.AttributeParameter(
                    name="b", type=ir.AttributeType.INT, required=True, default=None
                ),
                ir.schemas.AttributeParameter(
                    name="c",
                    type=ir.AttributeType.FLOAT,
                    required=False,
                    default=ir.Attr("c", ir.AttributeType.FLOAT, 100.0),
                ),
            ],
            outputs=[],
        )

        expected_inputs = [TEST_INPUT]
        expected_attributes = collections.OrderedDict(
            [
                ("b", 42),
                ("c", expected_c),
            ]
        )

        inputs, attributes = param_manipulation.separate_input_attributes_from_arguments(
            op_signature, args, kwargs
        )

        self.assertEqual(len(inputs), len(expected_inputs))
        for input_, expected_input in zip(inputs, expected_inputs):
            self.assertIs(input_, expected_input)
        self.assertEqual(attributes, expected_attributes)

    @parameterized.parameterized.expand(
        [
            (
                "extra_keyword",
                (TEST_INPUT, 42, 0.0),
                {"unknown": -1},
            ),
        ]
    )
    def test_it_raises_on_extra_args(self, _, args, kwargs):
        type_constraint = ir.schemas.TypeConstraintParam.any_tensor("T")
        op_signature = ir.schemas.OpSignature(
            domain="",
            name="TestOp",
            overload="",
            params=[
                ir.schemas.Parameter(
                    name="a", type_constraint=type_constraint, required=True, variadic=False
                ),
                ir.schemas.AttributeParameter(
                    name="b", type=ir.AttributeType.INT, required=True, default=None
                ),
                ir.schemas.AttributeParameter(
                    name="c",
                    type=ir.AttributeType.FLOAT,
                    required=False,
                    default=ir.Attr("c", ir.AttributeType.FLOAT, 100.0),
                ),
            ],
            outputs=[],
        )

        with self.assertRaises(TypeError):
            _, _ = param_manipulation.separate_input_attributes_from_arguments(
                op_signature, args, kwargs
            )

    @parameterized.parameterized.expand(
        [
            (True,),
            (False,),
        ]
    )
    def test_it_raises_on_extra_kwargs_when_not_allow_extra_kwargs(
        self,
        fill_defaults: bool,
    ):
        type_constraint = ir.schemas.TypeConstraintParam.any_tensor("T")
        op_signature = ir.schemas.OpSignature(
            domain="",
            name="TestOp",
            overload="",
            params=[
                ir.schemas.Parameter(
                    name="a", type_constraint=type_constraint, required=True, variadic=False
                ),
                ir.schemas.AttributeParameter(
                    name="b", type=ir.AttributeType.INT, required=True, default=None
                ),
                ir.schemas.AttributeParameter(
                    name="c",
                    type=ir.AttributeType.FLOAT,
                    required=False,
                    default=ir.Attr("c", ir.AttributeType.FLOAT, 100.0),
                ),
            ],
            outputs=[],
        )

        with self.assertRaises(TypeError):
            _, _ = param_manipulation.separate_input_attributes_from_arguments(
                op_signature,
                (TEST_INPUT, 42),
                {"c": 1.0, "extra": 42},
                fill_defaults=fill_defaults,
                allow_extra_kwargs=False,
            )

    @parameterized.parameterized.expand(
        [
            (True,),
            (False,),
        ]
    )
    def test_it_does_not_fill_default_when_fill_defaults_is_false(
        self, allow_extra_kwargs: bool
    ):
        type_constraint = ir.schemas.TypeConstraintParam.any_tensor("T")
        op_signature = ir.schemas.OpSignature(
            domain="",
            name="TestOp",
            overload="",
            params=[
                ir.schemas.Parameter(
                    name="a", type_constraint=type_constraint, required=True, variadic=False
                ),
                ir.schemas.AttributeParameter(
                    name="b", type=ir.AttributeType.INT, required=True, default=None
                ),
                ir.schemas.AttributeParameter(
                    name="c",
                    type=ir.AttributeType.FLOAT,
                    required=False,
                    default=ir.Attr("c", ir.AttributeType.FLOAT, 100.0),
                ),
            ],
            outputs=[],
        )

        inputs, attributes = param_manipulation.separate_input_attributes_from_arguments(
            op_signature,
            (TEST_INPUT, 42),
            {},
            fill_defaults=False,
            allow_extra_kwargs=allow_extra_kwargs,
        )

        self.assertEqual(inputs, [TEST_INPUT])
        self.assertEqual(attributes, collections.OrderedDict([("b", 42)]))

    @parameterized.parameterized.expand(
        [
            (True, True),
            (True, False),
            (False, True),
            (False, False),
        ]
    )
    def test_it_raises_on_insufficient_args(
        self, fill_defaults: bool, allow_extra_kwargs: bool
    ):
        type_constraint = ir.schemas.TypeConstraintParam.any_tensor("T")
        op_signature = ir.schemas.OpSignature(
            domain="",
            name="TestOp",
            overload="",
            params=[
                ir.schemas.Parameter(
                    name="a", type_constraint=type_constraint, required=True, variadic=False
                ),
                ir.schemas.AttributeParameter(
                    name="b", type=ir.AttributeType.INT, required=True, default=None
                ),
                ir.schemas.AttributeParameter(
                    name="c",
                    type=ir.AttributeType.FLOAT,
                    required=False,
                    default=ir.Attr("c", ir.AttributeType.FLOAT, 100.0),
                ),
            ],
            outputs=[],
        )

        with self.assertRaises(TypeError):
            _, _ = param_manipulation.separate_input_attributes_from_arguments(
                op_signature,
                (TEST_INPUT,),
                {},
                fill_defaults=fill_defaults,
                allow_extra_kwargs=allow_extra_kwargs,
            )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import unittest

import onnx
import onnx_ir as ir

import onnxscript.optimizer as optimizer


class OptimizerTest(unittest.TestCase):
    def test_folded_from_key_is_accessible(self):
        """Test that FOLDED_FROM_KEY is accessible from the public API."""
        self.assertTrue(hasattr(optimizer, "FOLDED_FROM_KEY"))
        self.assertIsInstance(optimizer.FOLDED_FROM_KEY, str)
        self.assertEqual(optimizer.FOLDED_FROM_KEY, "pkg.onnxscript.optimizer.folded_from")

    def _model_proto(self) -> onnx.ModelProto:
        return onnx.parser.parse_model(
            """
                <
                    ir_version: 8,
                    opset_import: ["pkg.onnxscript.torch_lib" : 1, "" : 18, "pkg.onnxscript.torch_lib.common" : 1],
                    producer_name: "pytorch",
                    producer_version: "2.2.0"
                >
                main_graph (float[3,5] l_tensor_x_) => (float[3,5] return_val)
                < _val_2, float[3,5] l_tensor_x_, float[2,5] getitem, float[1,5] getitem_1>
                {
                    _val_1 = Constant <value: tensor = int64 {2}> ()
                    _val_2 = pkg.onnxscript.torch_lib.aten_split <dim: int = 0> (l_tensor_x_, _val_1)
                    _val_3 = Constant <value: tensor = int64 {0}> ()
                    getitem = pkg.onnxscript.torch_lib.aten_getitem (_val_2, _val_3)
                    _val_5 = Constant <value: tensor = int64 {1}> ()
                    getitem_1 = pkg.onnxscript.torch_lib.aten_getitem (_val_2, _val_5)
                    return_val = Concat <axis: int = 0> (getitem_1, getitem)
                }

                <domain: "pkg.onnxscript.torch_lib", opset_import: ["" : 18]>
                aten_split (self, split_size) => (return_val)
                {
                    return_val = SplitToSequence <axis: int = @dim> (self, split_size)
                }

                <domain: "pkg.onnxscript.torch_lib", opset_import: ["" : 18]>
                aten_getitem (self, i) => (return_val)
                {
                    return_val = SequenceAt (self, i)
                }

                <domain: "pkg.onnxscript.torch_lib.common", opset_import: ["" : 18]>
                Rank (input) => (return_val)
                {
                    tmp = Shape (input)
                    return_val = Size (tmp)
                }

                <domain: "pkg.onnxscript.torch_lib.common", opset_import: ["" : 18]>
                IsScalar (input) => (return_val)
                {
                    tmp = Shape (input)
                    tmp_0 = Size (tmp)
                    tmp_1 = Constant <value_int: int = 0> ()
                    return_val = Equal (tmp_0, tmp_1)
                }
                """
        )

    def test_static_split_to_sequence_with_uneven_split_proto(self):
        model_proto = self._model_proto()
        optimized = optimizer.optimize(
            model_proto, num_iterations=1, onnx_shape_inference=False
        )
        self.assertEqual(len(optimized.graph.node), 2)
        self.assertEqual(len(optimized.graph.node[0].output), 2)
        self.assertEqual(optimized.graph.node[0].op_type, "Split")

    def test_static_split_to_sequence_with_uneven_split_ir(self):
        model_proto = self._model_proto()
        model_ir = ir.serde.deserialize_model(model_proto)
        optimizer.optimize_ir(model_ir, num_iterations=1, onnx_shape_inference=False)
        self.assertEqual(len(model_ir.graph), 2)
        self.assertEqual(len(model_ir.graph.node(0).outputs), 2)
        self.assertEqual(model_ir.graph.node(0).op_type, "Split")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import unittest

import numpy as np
import onnx
import onnx_ir as ir
from onnx_ir.passes.common import onnx_checker, shape_inference
from parameterized import parameterized

from onnxscript.rewriter import MatchingTracer, MatchStatus, RewriteRule, testing
from onnxscript.rewriter.rules.common._min_max_to_clip import (
    max_max_rule,
    max_min_rule,
    min_max_rule,
    min_min_rule,
    rules,
)


class _TestMinMaxToClipBase(unittest.TestCase):
    @property
    def rng(self):
        return np.random.default_rng(20250817)

    def clone_model(self, model: ir.Model) -> ir.Model:
        return ir.from_proto(ir.to_proto(model))

    def run_test(
        self,
        base_model: ir.Model,
        expected_op_types: list[str],
        dtype: str = "float",
    ):
        onnx_checker.CheckerPass(True)(base_model)
        base_model = shape_inference.infer_shapes(base_model)
        updated_model = self.clone_model(base_model)
        _ = rules.apply_to_model(updated_model)

        # Check expected op_types
        self.assertEqual([node.op_type for node in updated_model.graph], expected_op_types)

        # Check inference
        inputs = (
            self.rng.integers(
                low=-10,
                high=10,
                size=(2, *updated_model.graph.inputs[0].shape[1:]),
                dtype=np.int32,
            ),
        )
        if dtype == "float":
            inputs = (inputs[0].astype(np.float32),)

        testing.assert_numerically_equal(
            base_model,
            updated_model,
            inputs,
        )

        # Validate serialized model
        output_model_proto = ir.serde.serialize_model(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def run_failed_condition_test(
        self,
        base_model: ir.Model,
        rewrite_rule: RewriteRule,
        expected_message: str,
    ):
        onnx_checker.CheckerPass(True)(base_model)

        updated_model = self.clone_model(base_model)
        tracer = MatchingTracer()
        count = rewrite_rule.apply_to_model(updated_model, tracer=tracer)

        # Check that the model is unchanged
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[rewrite_rule][0]
        self.assertEqual(tracer_match.status.value, MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, expected_message)


class TestFuseSuccessiveMinOrMax(_TestMinMaxToClipBase):
    @parameterized.expand(
        [
            ("int32_min", "int32", "Min"),
            ("int32_max", "int32", "Max"),
            ("float32_min", "float", "Min"),
            ("float32_max", "float", "Max"),
        ]
    )
    def test_successful_fuse_successive_min_or_max(self, _, dtype, op_type):
        base_model = ir.from_onnx_text(f"""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model ({dtype}[N, 32, 14, 17] X) => ({dtype} [N, ?, ?, ?] Y)
            <{dtype}[1] cst1 = {{3}}, {dtype}[1] cst2 = {{6}}>
            {{
                x1 = {op_type}(X, cst1)
                Y = {op_type}(x1, cst2)
            }}
        """)
        self.run_test(base_model, expected_op_types=[op_type], dtype=dtype)

    @parameterized.expand(
        [
            ("int32_min_multi", "int32", "Min"),
            ("int32_max_multi", "int32", "Max"),
            ("float32_min_multi", "float", "Min"),
            ("float32_max_multi", "float", "Max"),
        ]
    )
    def test_successful_fuse_successive_min_or_max_multiple_inputs(self, _, dtype, op_type):
        base_model = ir.from_onnx_text(f"""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model ({dtype}[N, 3, 3] X) => ({dtype}[N, 3, 3] Y)
            <{dtype}[3] cst1 = {{2, 5, 8}},
             {dtype}[1] cst2 = {{4}},
             {dtype}[3] cst3 = {{3, 1, -6}},
             {dtype}[1] cst4 = {{10}},
             {dtype}[3] cst5 = {{-2, 7, 9}},
             {dtype}[1] cst6 = {{0}},
             {dtype}[3] cst7 = {{11, -3, 4}}>
            {{
                x1 = {op_type}(X, cst1, cst2, cst3, cst4)
                Y  = {op_type}(x1, cst5, cst6, cst7)
            }}
        """)
        self.run_test(base_model, expected_op_types=[op_type], dtype=dtype)

    @parameterized.expand(
        [
            ("int32_min", "Min"),
            ("int32_max", "Max"),
            ("float32_min", "Min"),
            ("float32_max", "Max"),
        ]
    )
    def test_successful_fuse_successive_min_or_max_constants(self, _, op_type):
        base_model = ir.from_onnx_text(f"""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] cst1 = {{3}}>
            {{
                x1 = {op_type}(X, cst1)
                cst2 = Constant<value_float=6.0>()
                Y = {op_type}(x1, cst2)
            }}
        """)
        self.run_test(base_model, expected_op_types=["Constant", op_type])

    @parameterized.expand(
        [
            ("min_nonconst", "Min", min_min_rule),
            ("max_nonconst", "Max", max_max_rule),
        ]
    )
    def test_failure_fuse_successive_min_or_max_non_constant(self, _, op_type, rewrite_rule):
        model = ir.from_onnx_text(f"""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float[N, ?, ?, ?] Y)
            <float[1] cst2 = {{6}}>
            {{
                cst1 = ReduceMean<keepdims=0>(X)
                x1 = {op_type}(X, cst1)
                Y = {op_type}(x1, cst2)
            }}
        """)
        self.run_failed_condition_test(model, rewrite_rule, "is not a constant.")

    @parameterized.expand(
        [
            ("min_graph_input", "Min"),
            ("max_graph_input", "Max"),
        ]
    )
    def test_successful_fuse_successive_min_or_max_graph_inputs_as_constants(self, _, op_type):
        base_model = ir.from_onnx_text(f"""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X, float[1] cst1, float[1] cst2) => (float[N, ?, ?, ?] Y)
            <float[1] cst1 = {{1.0}}, float[1] cst2 = {{6}}>
            {{
                x1 = {op_type}(X, cst1)
                Y = {op_type}(x1, cst2)
            }}
        """)
        self.run_test(base_model, expected_op_types=[op_type])


class TestMinMaxToClip(_TestMinMaxToClipBase):
    def test_successful_min_max_to_clip(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] min = {10.0}, float[1] max = {6.0}>
            {
                x1 = Min(X, min)
                Y = Max(x1, max)
            }
        """)
        self.run_test(base_model, expected_op_types=["Clip"])

    def test_successful_min_max_to_clip_constants(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] min = {10.0}>
            {
                x1 = Min(X, min)
                max = Constant<value_float=6.0>()
                Y = Max(x1, max)
            }
        """)
        self.run_test(base_model, expected_op_types=["Constant", "Clip"])

    def test_successful_min_max_to_clip_graph_inputs_as_constants(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X, float[1] min, float[1] max) => (float [N, ?, ?, ?] Y)
            <float[1] min = {12.0}, float[1] max = {6.0}>
            {
                x1 = Min(X, min)
                Y = Max(x1, max)
            }
        """)
        self.run_test(base_model, expected_op_types=["Clip"])

    def test_failure_min_max_to_clip_invalid_bounds(self):
        """Min node should have the max value and Max node should have the min value."""
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] min = {2.0}, float[1] max = {6.0}>
            {
                x1 = Min(X, min)
                Y = Max(x1, max)
            }
        """)
        self.run_failed_condition_test(base_model, min_max_rule, "Invalid bounds:")

    def test_failure_fuse_min_max_to_clip_non_constant(self):
        model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] max = {6.0}>
            {
                min = ReduceMean<keepdims=0>(X)
                x1 = Min(X, min)
                Y = Max(x1, max)
            }
        """)
        self.run_failed_condition_test(model, min_max_rule, "is not a constant.")

    def test_failure_min_max_to_clip_need_scalars(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 4, 4] X) => (float [N, ?, ?] Y)
            <float[4] min = {1.0, 2.0, -3.0, 0.0}, float[1] max = {6.0}>
            {
                x1 = Min(X, min)
                Y = Max(x1, max)
            }
        """)
        self.run_failed_condition_test(base_model, min_max_rule, "is not a scalar")


class TestMaxMinToClip(_TestMinMaxToClipBase):
    def test_successful_max_min_to_clip(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] min = {10.0}, float[1] max = {6.0}>
            {
                x1 = Max(X, max)
                Y = Min(x1, min)
            }
        """)
        self.run_test(base_model, expected_op_types=["Clip"])

    def test_successful_max_min_to_clip_constants(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] max = {6.0}>
            {
                x1 = Max(X, max)
                min = Constant<value_float=10.0>()
                Y = Min(x1, min)
            }
        """)
        self.run_test(base_model, expected_op_types=["Constant", "Clip"])

    def test_successful_max_min_to_clip_graph_inputs_as_constants(self):
        base_model = ir.from_onnx_text("""
                < ir_version: 10, opset_import: ["" : 20] >
                test_model (float[N, 32, 14, 17] X, float[1] min, float[1] max) => (float [N, ?, ?, ?] Y)
                <float[1] min = {12.0}, float[1] max = {6.0}>
                {
                    x1 = Max(X, max)
                    Y = Min(x1, min)
                }
            """)
        self.run_test(base_model, expected_op_types=["Clip"])

    def test_successful_max_min_to_clip_check_bounds(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] min = {2.0}, float[1] max = {6.0}>
            {
                x1 = Max(X, max)
                Y = Min(x1, min)
            }
        """)
        self.run_test(base_model, expected_op_types=["Clip"])

    def test_failure_fuse_max_min_to_clip_non_constant(self):
        model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14, 17] X) => (float [N, ?, ?, ?] Y)
            <float[1] max = {6.0}>
            {
                min = ReduceMean<keepdims=0>(X)
                x1 = Max(X, max)
                Y = Min(x1, min)
            }
        """)
        self.run_failed_condition_test(model, max_min_rule, "is not a constant.")

    def test_failure_max_min_to_clip_need_scalars(self):
        base_model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 4, 4] X) => (float [N, ?, ?] Y)
            <float[4] min = {1.0, 2.0, -3.0, 0.0}, float[1] max = {6.0}>
            {
                x1 = Max(X, min)
                Y = Min(x1, max)
            }
        """)
        self.run_failed_condition_test(base_model, max_min_rule, "is not a scalar")


class TestIntegrationMinMaxToClip(_TestMinMaxToClipBase):
    def test_successful_full_chain_fusion(self):
        model = ir.from_onnx_text("""
            < ir_version: 10, opset_import: ["" : 20] >
            test_model (float[N, 32, 14] X) => (float [N, ?, ?] Y)
            <float[1] min1 = {10.0}, float[1] min2 = {12.0}, float[1] min3 = {20.0}, float[1] min4 = {5.0},
            float[1] max1 = {6.0}, float[1] max2 = {8.0}, float[1] max3 = {5.0}>
            {
                x1 = Min(X, min1)
                x2 = Min(x1, min2)
                x3 = Max(x2, max1)
                x4 = Max(x3, max2)
                x5 = Min(x4, min3)
                x6 = Max(x5, max3)
                Y = Min(x6, min4)
            }
        """)
        self.run_test(model, expected_op_types=["Clip", "Clip", "Clip"])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

# Generated from Phi4LM 2 Layer ONNX model produced by the new (Dynamo) exporter
# ruff: noqa: F821

import numpy
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18
from onnxscript.onnx_types import BOOL, FLOAT, INT64

value_infos = {
    "model_embed_tokens_weight": FLOAT[100352, 5120],
    "model_layers_0_self_attn_o_proj_weight": FLOAT[5120, 5120],
    "model_layers_0_self_attn_qkv_proj_weight": FLOAT[7680, 5120],
    "model_layers_0_mlp_gate_up_proj_weight": FLOAT[35840, 5120],
    "model_layers_0_mlp_down_proj_weight": FLOAT[5120, 17920],
    "model_layers_0_input_layernorm_weight": FLOAT[5120],
    "model_layers_0_post_attention_layernorm_weight": FLOAT[5120],
    "model_layers_1_self_attn_o_proj_weight": FLOAT[5120, 5120],
    "model_layers_1_self_attn_qkv_proj_weight": FLOAT[7680, 5120],
    "model_layers_1_mlp_gate_up_proj_weight": FLOAT[35840, 5120],
    "model_layers_1_mlp_down_proj_weight": FLOAT[5120, 17920],
    "model_layers_1_input_layernorm_weight": FLOAT[5120],
    "model_layers_1_post_attention_layernorm_weight": FLOAT[5120],
    "model_norm_weight": FLOAT[5120],
    "lm_head_weight": FLOAT[100352, 5120],
    "expand_2": FLOAT[1, 64, 1],
    "val_1": INT64[1],
    "sym_size_int_61": INT64,
    "val_5": INT64[1],
    "sym_size_int_67": INT64,
    "val_6": INT64[1],
    "embedding": FLOAT["s34", "s16", 5120],
    "add_4": INT64,
    "val_11": INT64,
    "arange": INT64["s16"],
    "val_12": INT64[1],
    "unsqueeze": INT64[1, "s16"],
    "val_14": FLOAT,
    "val_17": INT64[1],
    "val_18": INT64[1],
    "val_19": INT64[2],
    "full": FLOAT["s16", "s16 + s17"],
    "val_22": INT64,
    "val_23": INT64,
    "arange_1": INT64["s16 + s17"],
    "val_25": INT64[2],
    "view": INT64["s16", 1],
    "gt": BOOL["s16", "s16 + s17"],
    "convert_element_type_default": FLOAT["s16", "s16 + s17"],
    "mul_14": FLOAT["s16", "s16 + s17"],
    "val_26": INT64[1],
    "val_805": INT64[2],
    "unsqueeze_4": FLOAT[1, 1, "s16", "s16 + s17"],
    "val_27": INT64,
    "val_35": INT64,
    "val_53": INT64[1],
    "val_54": INT64[4],
    "val_56": INT64[4],
    "expand_1": FLOAT["s34", 1, "s16", "s16 + s17"],
    "val_65": INT64,
    "val_76": INT64[1],
    "val_78": INT64[1],
    "val_79": INT64[1],
    "val_82": INT64[1],
    "val_83": INT64[1],
    "slice_8": FLOAT["s34", 1, "s16", "s16 + s17"],
    "val_94": INT64[1],
    "val_806": INT64[2],
    "unsqueeze_6": INT64["s34", 1, 1, "s16 + s17"],
    "convert_element_type_default_1": FLOAT["s34", 1, 1, "s16 + s17"],
    "add_86": FLOAT["s34", 1, "s16", "s16 + s17"],
    "scalar_tensor_default": FLOAT,
    "eq_65": BOOL["s34", 1, "s16", "s16 + s17"],
    "val_123": INT64[1],
    "val_125": INT64[1],
    "val_126": INT64[1],
    "val_129": INT64[1],
    "val_130": INT64[1],
    "slice_14": FLOAT["s34", 1, "s16", "s16 + s17"],
    "val_131": FLOAT,
    "masked_fill": FLOAT["s34", 1, "s16", "s16 + s17"],
    "val_183": INT64[4],
    "val_184": INT64,
    "val_185": INT64[None],
    "val_190": INT64[None, 1],
    "val_191": FLOAT["s16", 1, "s34", "s16 + s17"],
    "val_192": FLOAT["s16", 1, "s34", "s16 + s17"],
    "val_193": FLOAT["s16", 1, "s34", "s16 + s17"],
    "val_195": INT64[4],
    "val_196": INT64,
    "val_197": INT64[None],
    "val_202": INT64[None, 1],
    "val_203": FLOAT[1, "s34", "s16", "s16 + s17"],
    "val_204": FLOAT[1, "s34", "s16", "s16 + s17"],
    "val_205": FLOAT[1, "s34", "s16", "s16 + s17"],
    "slice_scatter_1": FLOAT["s34", 1, "s16", "s16 + s17"],
    "val_207": INT64[4],
    "val_208": INT64,
    "val_209": INT64[None],
    "val_214": INT64[None, 1],
    "slice_scatter_2": FLOAT["s34", 1, "s16", "s16 + s17"],
    "unsqueeze_9": INT64[1, 1, "s16"],
    "_to_copy": FLOAT[1, 1, "s16"],
    "matmul": FLOAT[1, 64, "s16"],
    "transpose": FLOAT[1, "s16", 64],
    "cat": FLOAT[1, "s16", 128],
    "cos": FLOAT[1, "s16", 128],
    "sin": FLOAT[1, "s16", 128],
    "val_248": FLOAT,
    "pow_1": FLOAT["s34", "s16", 5120],
    "val_250": INT64[1],
    "mean": FLOAT["s34", "s16", 1],
    "val_251": FLOAT,
    "add_189": FLOAT["s34", "s16", 1],
    "val_252": FLOAT["s34", "s16", 1],
    "rsqrt": FLOAT["s34", "s16", 1],
    "mul_167": FLOAT["s34", "s16", 5120],
    "mul_171": FLOAT["s34", "s16", 5120],
    "val_253": FLOAT[5120, 7680],
    "linear": FLOAT["s34", "s16", 7680],
    "val_256": INT64[1],
    "val_260": INT64[1],
    "val_263": INT64[1],
    "val_264": INT64[1],
    "slice_26": FLOAT["s34", "s16", 5120],
    "val_267": INT64[1],
    "val_271": INT64[1],
    "val_274": INT64[1],
    "val_275": INT64[1],
    "slice_27": FLOAT["s34", "s16", 1280],
    "val_278": INT64[1],
    "val_281": INT64[1],
    "val_284": INT64[1],
    "val_285": INT64[1],
    "slice_28": FLOAT["s34", "s16", 1280],
    "val_290": INT64[1],
    "val_291": INT64[4],
    "view_1": FLOAT["s34", "s16", 40, 128],
    "transpose_1": FLOAT["s34", 40, "s16", 128],
    "val_297": INT64[4],
    "view_2": FLOAT["s34", "s16", 10, 128],
    "transpose_2": FLOAT["s34", 10, "s16", 128],
    "val_303": INT64[4],
    "view_3": FLOAT["s34", "s16", 10, 128],
    "transpose_3": FLOAT["s34", 10, "s16", 128],
    "unsqueeze_10": FLOAT[1, 1, "s16", 128],
    "unsqueeze_11": FLOAT[1, 1, "s16", 128],
    "mul_223": FLOAT["s34", 40, "s16", 128],
    "val_328": INT64[1],
    "val_332": INT64[1],
    "val_335": INT64[1],
    "val_336": INT64[1],
    "slice_31": FLOAT["s34", 40, "s16", 64],
    "val_339": INT64[1],
    "val_342": INT64[1],
    "val_345": INT64[1],
    "val_346": INT64[1],
    "slice_32": FLOAT["s34", 40, "s16", 64],
    "neg": FLOAT["s34", 40, "s16", 64],
    "cat_1": FLOAT["s34", 40, "s16", 128],
    "mul_240": FLOAT["s34", 40, "s16", 128],
    "add_304": FLOAT["s34", 40, "s16", 128],
    "mul_252": FLOAT["s34", 10, "s16", 128],
    "val_349": INT64[1],
    "val_352": INT64[1],
    "val_355": INT64[1],
    "val_356": INT64[1],
    "slice_33": FLOAT["s34", 10, "s16", 64],
    "val_359": INT64[1],
    "val_362": INT64[1],
    "val_365": INT64[1],
    "val_366": INT64[1],
    "slice_34": FLOAT["s34", 10, "s16", 64],
    "neg_1": FLOAT["s34", 10, "s16", 64],
    "cat_3": FLOAT["s34", 10, "s16", 128],
    "mul_269": FLOAT["s34", 10, "s16", 128],
    "add_345": FLOAT["s34", 10, "s16", 128],
    "unsqueeze_12": FLOAT["s34", 10, 1, "s16 + s17", 128],
    "val_410": INT64[1],
    "val_411": INT64[1],
    "val_412": INT64[1],
    "val_413": INT64[1],
    "val_414": INT64[5],
    "val_416": INT64[5],
    "expand_3": FLOAT["s34", 10, 4, "s16 + s17", 128],
    "val_419": INT64[1],
    "val_420": INT64[1],
    "val_421": INT64[1],
    "val_422": INT64[4],
    "_unsafe_view": FLOAT["s34", 40, "s16 + s17", 128],
    "unsqueeze_13": FLOAT["s34", 10, 1, "s16 + s17", 128],
    "val_466": INT64[1],
    "val_467": INT64[1],
    "val_468": INT64[5],
    "val_470": INT64[5],
    "expand_4": FLOAT["s34", 10, 4, "s16 + s17", 128],
    "val_473": INT64[1],
    "val_474": INT64[1],
    "val_475": INT64[4],
    "_unsafe_view_1": FLOAT["s34", 40, "s16 + s17", 128],
    "transpose_4": FLOAT["s34", 40, 128, "s16 + s17"],
    "matmul_1": FLOAT["s34", 40, "s16", "s16 + s17"],
    "val_477": FLOAT,
    "mul_433": FLOAT["s34", 40, "s16", "s16 + s17"],
    "val_496": INT64[1],
    "val_498": INT64[1],
    "val_499": INT64[1],
    "val_502": INT64[1],
    "val_503": INT64[1],
    "slice_50": FLOAT["s34", 1, "s16", "s16 + s17"],
    "add_491": FLOAT["s34", 40, "s16", "s16 + s17"],
    "val_504": FLOAT["s34", 40, "s16", "s16 + s17"],
    "matmul_2": FLOAT["s34", 40, "s16", 128],
    "transpose_5": FLOAT["s34", "s16", 40, 128],
    "val_509": INT64[3],
    "view_4": FLOAT["s34", "s16", 5120],
    "val_511": FLOAT[5120, 5120],
    "linear_1": FLOAT["s34", "s16", 5120],
    "add_534": FLOAT["s34", "s16", 5120],
    "val_512": FLOAT,
    "pow_2": FLOAT["s34", "s16", 5120],
    "val_514": INT64[1],
    "mean_1": FLOAT["s34", "s16", 1],
    "add_547": FLOAT["s34", "s16", 1],
    "val_515": FLOAT["s34", "s16", 1],
    "rsqrt_1": FLOAT["s34", "s16", 1],
    "mul_506": FLOAT["s34", "s16", 5120],
    "mul_510": FLOAT["s34", "s16", 5120],
    "val_516": FLOAT[5120, 35840],
    "linear_2": FLOAT["s34", "s16", 35840],
    "split_split_0": FLOAT["s34", "s16", 17920],
    "split_split_1": FLOAT["s34", "s16", 17920],
    "val_518": FLOAT["s34", "s16", 17920],
    "silu": FLOAT["s34", "s16", 17920],
    "mul_526": FLOAT["s34", "s16", 17920],
    "val_519": FLOAT[17920, 5120],
    "linear_3": FLOAT["s34", "s16", 5120],
    "add_592": FLOAT["s34", "s16", 5120],
    "val_520": FLOAT,
    "pow_3": FLOAT["s34", "s16", 5120],
    "val_522": INT64[1],
    "mean_2": FLOAT["s34", "s16", 1],
    "add_605": FLOAT["s34", "s16", 1],
    "val_523": FLOAT["s34", "s16", 1],
    "rsqrt_2": FLOAT["s34", "s16", 1],
    "mul_548": FLOAT["s34", "s16", 5120],
    "mul_552": FLOAT["s34", "s16", 5120],
    "val_524": FLOAT[5120, 7680],
    "linear_4": FLOAT["s34", "s16", 7680],
    "val_527": INT64[1],
    "val_530": INT64[1],
    "val_533": INT64[1],
    "val_534": INT64[1],
    "slice_51": FLOAT["s34", "s16", 5120],
    "val_537": INT64[1],
    "val_540": INT64[1],
    "val_543": INT64[1],
    "val_544": INT64[1],
    "slice_52": FLOAT["s34", "s16", 1280],
    "val_547": INT64[1],
    "val_550": INT64[1],
    "val_553": INT64[1],
    "val_554": INT64[1],
    "slice_53": FLOAT["s34", "s16", 1280],
    "val_559": INT64[4],
    "view_5": FLOAT["s34", "s16", 40, 128],
    "transpose_6": FLOAT["s34", 40, "s16", 128],
    "val_565": INT64[4],
    "view_6": FLOAT["s34", "s16", 10, 128],
    "transpose_7": FLOAT["s34", 10, "s16", 128],
    "val_571": INT64[4],
    "view_7": FLOAT["s34", "s16", 10, 128],
    "transpose_8": FLOAT["s34", 10, "s16", 128],
    "unsqueeze_14": FLOAT[1, 1, "s16", 128],
    "unsqueeze_15": FLOAT[1, 1, "s16", 128],
    "mul_604": FLOAT["s34", 40, "s16", 128],
    "val_595": INT64[1],
    "val_598": INT64[1],
    "val_601": INT64[1],
    "val_602": INT64[1],
    "slice_56": FLOAT["s34", 40, "s16", 64],
    "val_605": INT64[1],
    "val_608": INT64[1],
    "val_611": INT64[1],
    "val_612": INT64[1],
    "slice_57": FLOAT["s34", 40, "s16", 64],
    "neg_2": FLOAT["s34", 40, "s16", 64],
    "cat_7": FLOAT["s34", 40, "s16", 128],
    "mul_621": FLOAT["s34", 40, "s16", 128],
    "add_720": FLOAT["s34", 40, "s16", 128],
    "mul_633": FLOAT["s34", 10, "s16", 128],
    "val_615": INT64[1],
    "val_618": INT64[1],
    "val_621": INT64[1],
    "val_622": INT64[1],
    "slice_58": FLOAT["s34", 10, "s16", 64],
    "val_625": INT64[1],
    "val_628": INT64[1],
    "val_631": INT64[1],
    "val_632": INT64[1],
    "slice_59": FLOAT["s34", 10, "s16", 64],
    "neg_3": FLOAT["s34", 10, "s16", 64],
    "cat_9": FLOAT["s34", 10, "s16", 128],
    "mul_650": FLOAT["s34", 10, "s16", 128],
    "add_761": FLOAT["s34", 10, "s16", 128],
    "unsqueeze_16": FLOAT["s34", 10, 1, "s16 + s17", 128],
    "val_675": INT64[1],
    "val_676": INT64[1],
    "val_677": INT64[5],
    "val_679": INT64[5],
    "expand_5": FLOAT["s34", 10, 4, "s16 + s17", 128],
    "val_682": INT64[1],
    "val_683": INT64[1],
    "val_684": INT64[4],
    "_unsafe_view_2": FLOAT["s34", 40, "s16 + s17", 128],
    "unsqueeze_17": FLOAT["s34", 10, 1, "s16 + s17", 128],
    "val_728": INT64[1],
    "val_729": INT64[1],
    "val_730": INT64[5],
    "val_732": INT64[5],
    "expand_6": FLOAT["s34", 10, 4, "s16 + s17", 128],
    "val_735": INT64[1],
    "val_736": INT64[1],
    "val_737": INT64[4],
    "_unsafe_view_3": FLOAT["s34", 40, "s16 + s17", 128],
    "transpose_9": FLOAT["s34", 40, 128, "s16 + s17"],
    "matmul_3": FLOAT["s34", 40, "s16", "s16 + s17"],
    "mul_814": FLOAT["s34", 40, "s16", "s16 + s17"],
    "val_757": INT64[1],
    "val_759": INT64[1],
    "val_760": INT64[1],
    "val_763": INT64[1],
    "val_764": INT64[1],
    "slice_75": FLOAT["s34", 1, "s16", "s16 + s17"],
    "add_907": FLOAT["s34", 40, "s16", "s16 + s17"],
    "val_765": FLOAT["s34", 40, "s16", "s16 + s17"],
    "matmul_4": FLOAT["s34", 40, "s16", 128],
    "transpose_10": FLOAT["s34", "s16", 40, 128],
    "val_770": INT64[3],
    "view_8": FLOAT["s34", "s16", 5120],
    "val_772": FLOAT[5120, 5120],
    "linear_5": FLOAT["s34", "s16", 5120],
    "add_950": FLOAT["s34", "s16", 5120],
    "val_773": FLOAT,
    "pow_4": FLOAT["s34", "s16", 5120],
    "val_775": INT64[1],
    "mean_3": FLOAT["s34", "s16", 1],
    "add_963": FLOAT["s34", "s16", 1],
    "val_776": FLOAT["s34", "s16", 1],
    "rsqrt_3": FLOAT["s34", "s16", 1],
    "mul_887": FLOAT["s34", "s16", 5120],
    "mul_891": FLOAT["s34", "s16", 5120],
    "val_777": FLOAT[5120, 35840],
    "linear_6": FLOAT["s34", "s16", 35840],
    "split_1_split_0": FLOAT["s34", "s16", 17920],
    "split_1_split_1": FLOAT["s34", "s16", 17920],
    "val_778": FLOAT["s34", "s16", 17920],
    "silu_1": FLOAT["s34", "s16", 17920],
    "mul_907": FLOAT["s34", "s16", 17920],
    "val_779": FLOAT[17920, 5120],
    "linear_7": FLOAT["s34", "s16", 5120],
    "add_1008": FLOAT["s34", "s16", 5120],
    "val_780": FLOAT,
    "pow_5": FLOAT["s34", "s16", 5120],
    "val_782": INT64[1],
    "mean_4": FLOAT["s34", "s16", 1],
    "add_1021": FLOAT["s34", "s16", 1],
    "val_783": FLOAT["s34", "s16", 1],
    "rsqrt_4": FLOAT["s34", "s16", 1],
    "mul_929": FLOAT["s34", "s16", 5120],
    "mul_933": FLOAT["s34", "s16", 5120],
    "val_804": FLOAT[5120, 100352],
}


def make_model(
    model_embed_tokens_weight,
    model_layers_0_self_attn_o_proj_weight,
    model_layers_0_self_attn_qkv_proj_weight,
    model_layers_0_mlp_gate_up_proj_weight,
    model_layers_0_mlp_down_proj_weight,
    model_layers_0_input_layernorm_weight,
    model_layers_0_post_attention_layernorm_weight,
    model_layers_1_self_attn_o_proj_weight,
    model_layers_1_self_attn_qkv_proj_weight,
    model_layers_1_mlp_gate_up_proj_weight,
    model_layers_1_mlp_down_proj_weight,
    model_layers_1_input_layernorm_weight,
    model_layers_1_post_attention_layernorm_weight,
    model_norm_weight,
    lm_head_weight,
    expand_2,
):
    @script()
    def main_graph(
        input_ids: INT64["s34", "s16"],
        attention_mask: INT64["s34", "s16 + s17"],
        past_key_values_key_cache_0: FLOAT["s34", 10, "s17", 128],
        past_key_values_key_cache_1: FLOAT["s34", 10, "s17", 128],
        past_key_values_value_cache_0: FLOAT["s34", 10, "s17", 128],
        past_key_values_value_cache_1: FLOAT["s34", 10, "s17", 128],
    ) -> (
        FLOAT["s34", "s16", 100352],
        FLOAT["s34", 10, "s16 + s17", 128],
        FLOAT["s34", 10, "s16 + s17", 128],
        FLOAT["s34", 10, "s16 + s17", 128],
        FLOAT["s34", 10, "s16 + s17", 128],
    ):
        val_1 = opset18.Shape(input_ids, end=2, start=1)
        sym_size_int_61 = opset18.Squeeze(val_1)
        val_5 = opset18.Shape(past_key_values_key_cache_1, end=3, start=2)
        sym_size_int_67 = opset18.Squeeze(val_5)
        val_6 = opset18.Shape(past_key_values_value_cache_0, end=1, start=0)
        embedding = opset18.Gather(model_embed_tokens_weight, input_ids, axis=0)
        add_4 = opset18.Add(sym_size_int_67, sym_size_int_61)
        arange = opset18.Range(sym_size_int_67, add_4, 1)
        unsqueeze = opset18.Unsqueeze(arange, [0])
        val_18 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_19 = opset18.Concat(val_1, val_18, axis=0)
        full = opset18.Expand(-3.4028235e38, val_19)
        arange_1 = opset18.Range(0, add_4, 1)
        view = opset18.Reshape(arange, [-1, 1], allowzero=1)
        gt = opset18.Greater(arange_1, view)
        convert_element_type_default = opset18.Cast(gt, to=1)
        mul_14 = opset18.Mul(full, convert_element_type_default)
        unsqueeze_4 = opset18.Unsqueeze(mul_14, [0, 1])
        val_54 = opset18.Concat(val_6, [1], [-1], [-1], axis=0)
        val_56 = opset18.Abs(val_54)
        expand_1 = opset18.Expand(unsqueeze_4, val_56)
        val_76 = opset18.Constant(value_ints=[0])
        val_78 = opset18.Constant(value_ints=[-1])
        val_79 = opset18.Reshape(add_4, val_78, allowzero=0)
        val_83 = opset18.Constant(value_ints=[1])
        slice_8 = opset18.Slice(expand_1, val_76, val_79, [3], val_83)
        unsqueeze_6 = opset18.Unsqueeze(attention_mask, [1, 2])
        convert_element_type_default_1 = opset18.Cast(unsqueeze_6, to=1)
        add_86 = opset18.Add(slice_8, convert_element_type_default_1)
        eq_65 = opset18.Equal(add_86, 0.0)
        val_123 = opset18.Constant(value_ints=[0])
        val_125 = opset18.Constant(value_ints=[-1])
        val_126 = opset18.Reshape(add_4, val_125, allowzero=0)
        val_130 = opset18.Constant(value_ints=[1])
        slice_14 = opset18.Slice(expand_1, val_123, val_126, [3], val_130)
        masked_fill = opset18.Where(eq_65, -3.4028235e38, slice_14)
        val_183 = opset18.Shape(expand_1, start=0)
        val_184 = opset18.Gather(val_183, 2, axis=0)
        val_185 = opset18.Range(0, val_184, 1)
        val_190 = opset18.Unsqueeze(val_185, [-1])
        val_191 = opset18.Transpose(masked_fill, perm=[2, 1, 0, 3])
        val_192 = opset18.Transpose(expand_1, perm=[2, 1, 0, 3])
        val_193 = opset18.ScatterND(val_192, val_190, val_191, reduction="none")
        val_195 = opset18.Shape(expand_1, start=0)
        val_196 = opset18.Gather(val_195, 1, axis=0)
        val_197 = opset18.Range(0, val_196, 1)
        val_202 = opset18.Unsqueeze(val_197, [-1])
        val_203 = opset18.Transpose(val_193, perm=[1, 2, 0, 3])
        val_204 = opset18.Transpose(expand_1, perm=[1, 0, 2, 3])
        val_205 = opset18.ScatterND(val_204, val_202, val_203, reduction="none")
        slice_scatter_1 = opset18.Transpose(val_205, perm=[1, 0, 2, 3])
        val_207 = opset18.Shape(expand_1, start=0)
        val_208 = opset18.Gather(val_207, 0, axis=0)
        val_209 = opset18.Range(0, val_208, 1)
        val_214 = opset18.Unsqueeze(val_209, [-1])
        slice_scatter_2 = opset18.ScatterND(
            expand_1, val_214, slice_scatter_1, reduction="none"
        )
        unsqueeze_9 = opset18.Unsqueeze(unsqueeze, [1])
        _to_copy = opset18.Cast(unsqueeze_9, to=1)
        matmul = opset18.MatMul(expand_2, _to_copy)
        transpose = opset18.Transpose(matmul, perm=[0, 2, 1])
        cat = opset18.Concat(transpose, transpose, axis=-1)
        cos = opset18.Cos(cat)
        sin = opset18.Sin(cat)
        pow_1 = opset18.Pow(embedding, 2.0)
        mean = opset18.ReduceMean(pow_1, [-1], noop_with_empty_axes=0, keepdims=1)
        add_189 = opset18.Add(mean, 1e-05)
        val_252 = opset18.Sqrt(add_189)
        rsqrt = opset18.Reciprocal(val_252)
        mul_167 = opset18.Mul(embedding, rsqrt)
        mul_171 = opset18.Mul(model_layers_0_input_layernorm_weight, mul_167)
        val_253 = opset18.Transpose(model_layers_0_self_attn_qkv_proj_weight, perm=[1, 0])
        linear = opset18.MatMul(mul_171, val_253)
        val_264 = opset18.Constant(value_ints=[1])
        slice_26 = opset18.Slice(linear, [0], [5120], [2], val_264)
        val_275 = opset18.Constant(value_ints=[1])
        slice_27 = opset18.Slice(linear, [5120], [6400], [2], val_275)
        val_285 = opset18.Constant(value_ints=[1])
        slice_28 = opset18.Slice(linear, [6400], [9223372036854775807], [2], val_285)
        val_291 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_1 = opset18.Reshape(slice_26, val_291, allowzero=1)
        transpose_1 = opset18.Transpose(view_1, perm=[0, 2, 1, 3])
        val_297 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_2 = opset18.Reshape(slice_27, val_297, allowzero=1)
        transpose_2 = opset18.Transpose(view_2, perm=[0, 2, 1, 3])
        val_303 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_3 = opset18.Reshape(slice_28, val_303, allowzero=1)
        transpose_3 = opset18.Transpose(view_3, perm=[0, 2, 1, 3])
        unsqueeze_10 = opset18.Unsqueeze(cos, [1])
        unsqueeze_11 = opset18.Unsqueeze(sin, [1])
        mul_223 = opset18.Mul(transpose_1, unsqueeze_10)
        val_336 = opset18.Constant(value_ints=[1])
        slice_31 = opset18.Slice(transpose_1, [0], [64], [3], val_336)
        val_346 = opset18.Constant(value_ints=[1])
        slice_32 = opset18.Slice(transpose_1, [64], [9223372036854775807], [3], val_346)
        neg = opset18.Neg(slice_32)
        cat_1 = opset18.Concat(neg, slice_31, axis=-1)
        mul_240 = opset18.Mul(cat_1, unsqueeze_11)
        add_304 = opset18.Add(mul_223, mul_240)
        mul_252 = opset18.Mul(transpose_2, unsqueeze_10)
        val_356 = opset18.Constant(value_ints=[1])
        slice_33 = opset18.Slice(transpose_2, [0], [64], [3], val_356)
        val_366 = opset18.Constant(value_ints=[1])
        slice_34 = opset18.Slice(transpose_2, [64], [9223372036854775807], [3], val_366)
        neg_1 = opset18.Neg(slice_34)
        cat_3 = opset18.Concat(neg_1, slice_33, axis=-1)
        mul_269 = opset18.Mul(cat_3, unsqueeze_11)
        add_345 = opset18.Add(mul_252, mul_269)
        cat_5 = opset18.Concat(past_key_values_key_cache_0, add_345, axis=-2)
        cat_6 = opset18.Concat(past_key_values_value_cache_0, transpose_3, axis=-2)
        unsqueeze_12 = opset18.Unsqueeze(cat_5, [2])
        val_413 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_414 = opset18.Concat(val_6, [10], [4], val_413, [128], axis=0)
        val_416 = opset18.Abs(val_414)
        expand_3 = opset18.Expand(unsqueeze_12, val_416)
        val_421 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_422 = opset18.Concat(val_6, [40], val_421, [128], axis=0)
        _unsafe_view = opset18.Reshape(expand_3, val_422, allowzero=1)
        unsqueeze_13 = opset18.Unsqueeze(cat_6, [2])
        val_467 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_468 = opset18.Concat(val_6, [10], [4], val_467, [128], axis=0)
        val_470 = opset18.Abs(val_468)
        expand_4 = opset18.Expand(unsqueeze_13, val_470)
        val_474 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_475 = opset18.Concat(val_6, [40], val_474, [128], axis=0)
        _unsafe_view_1 = opset18.Reshape(expand_4, val_475, allowzero=1)
        transpose_4 = opset18.Transpose(_unsafe_view, perm=[0, 1, 3, 2])
        matmul_1 = opset18.MatMul(add_304, transpose_4)
        mul_433 = opset18.Mul(matmul_1, 0.088388346)
        val_496 = opset18.Constant(value_ints=[0])
        val_498 = opset18.Constant(value_ints=[-1])
        val_499 = opset18.Reshape(add_4, val_498, allowzero=0)
        val_503 = opset18.Constant(value_ints=[1])
        slice_50 = opset18.Slice(slice_scatter_2, val_496, val_499, [3], val_503)
        add_491 = opset18.Add(mul_433, slice_50)
        val_504 = opset18.Softmax(add_491, axis=-1)
        matmul_2 = opset18.MatMul(val_504, _unsafe_view_1)
        transpose_5 = opset18.Transpose(matmul_2, perm=[0, 2, 1, 3])
        val_509 = opset18.Concat(val_6, val_1, [-1], axis=0)
        view_4 = opset18.Reshape(transpose_5, val_509, allowzero=1)
        val_511 = opset18.Transpose(model_layers_0_self_attn_o_proj_weight, perm=[1, 0])
        linear_1 = opset18.MatMul(view_4, val_511)
        add_534 = opset18.Add(embedding, linear_1)
        pow_2 = opset18.Pow(add_534, 2.0)
        mean_1 = opset18.ReduceMean(pow_2, [-1], noop_with_empty_axes=0, keepdims=1)
        add_547 = opset18.Add(mean_1, 1e-05)
        val_515 = opset18.Sqrt(add_547)
        rsqrt_1 = opset18.Reciprocal(val_515)
        mul_506 = opset18.Mul(add_534, rsqrt_1)
        mul_510 = opset18.Mul(model_layers_0_post_attention_layernorm_weight, mul_506)
        val_516 = opset18.Transpose(model_layers_0_mlp_gate_up_proj_weight, perm=[1, 0])
        linear_2 = opset18.MatMul(mul_510, val_516)
        split_split_0, split_split_1 = opset18.Split(linear_2, axis=2, num_outputs=2)
        val_518 = opset18.Sigmoid(split_split_0)
        silu = opset18.Mul(split_split_0, val_518)
        mul_526 = opset18.Mul(split_split_1, silu)
        val_519 = opset18.Transpose(model_layers_0_mlp_down_proj_weight, perm=[1, 0])
        linear_3 = opset18.MatMul(mul_526, val_519)
        add_592 = opset18.Add(add_534, linear_3)
        pow_3 = opset18.Pow(add_592, 2.0)
        mean_2 = opset18.ReduceMean(pow_3, [-1], noop_with_empty_axes=0, keepdims=1)
        add_605 = opset18.Add(mean_2, 1e-05)
        val_523 = opset18.Sqrt(add_605)
        rsqrt_2 = opset18.Reciprocal(val_523)
        mul_548 = opset18.Mul(add_592, rsqrt_2)
        mul_552 = opset18.Mul(model_layers_1_input_layernorm_weight, mul_548)
        val_524 = opset18.Transpose(model_layers_1_self_attn_qkv_proj_weight, perm=[1, 0])
        linear_4 = opset18.MatMul(mul_552, val_524)
        val_534 = opset18.Constant(value_ints=[1])
        slice_51 = opset18.Slice(linear_4, [0], [5120], [2], val_534)
        val_544 = opset18.Constant(value_ints=[1])
        slice_52 = opset18.Slice(linear_4, [5120], [6400], [2], val_544)
        val_554 = opset18.Constant(value_ints=[1])
        slice_53 = opset18.Slice(linear_4, [6400], [9223372036854775807], [2], val_554)
        val_559 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_5 = opset18.Reshape(slice_51, val_559, allowzero=1)
        transpose_6 = opset18.Transpose(view_5, perm=[0, 2, 1, 3])
        val_565 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_6 = opset18.Reshape(slice_52, val_565, allowzero=1)
        transpose_7 = opset18.Transpose(view_6, perm=[0, 2, 1, 3])
        val_571 = opset18.Concat(val_6, val_1, [-1], [128], axis=0)
        view_7 = opset18.Reshape(slice_53, val_571, allowzero=1)
        transpose_8 = opset18.Transpose(view_7, perm=[0, 2, 1, 3])
        unsqueeze_14 = opset18.Unsqueeze(cos, [1])
        unsqueeze_15 = opset18.Unsqueeze(sin, [1])
        mul_604 = opset18.Mul(transpose_6, unsqueeze_14)
        val_602 = opset18.Constant(value_ints=[1])
        slice_56 = opset18.Slice(transpose_6, [0], [64], [3], val_602)
        val_612 = opset18.Constant(value_ints=[1])
        slice_57 = opset18.Slice(transpose_6, [64], [9223372036854775807], [3], val_612)
        neg_2 = opset18.Neg(slice_57)
        cat_7 = opset18.Concat(neg_2, slice_56, axis=-1)
        mul_621 = opset18.Mul(cat_7, unsqueeze_15)
        add_720 = opset18.Add(mul_604, mul_621)
        mul_633 = opset18.Mul(transpose_7, unsqueeze_14)
        val_622 = opset18.Constant(value_ints=[1])
        slice_58 = opset18.Slice(transpose_7, [0], [64], [3], val_622)
        val_632 = opset18.Constant(value_ints=[1])
        slice_59 = opset18.Slice(transpose_7, [64], [9223372036854775807], [3], val_632)
        neg_3 = opset18.Neg(slice_59)
        cat_9 = opset18.Concat(neg_3, slice_58, axis=-1)
        mul_650 = opset18.Mul(cat_9, unsqueeze_15)
        add_761 = opset18.Add(mul_633, mul_650)
        cat_11 = opset18.Concat(past_key_values_key_cache_1, add_761, axis=-2)
        cat_12 = opset18.Concat(past_key_values_value_cache_1, transpose_8, axis=-2)
        unsqueeze_16 = opset18.Unsqueeze(cat_11, [2])
        val_676 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_677 = opset18.Concat(val_6, [10], [4], val_676, [128], axis=0)
        val_679 = opset18.Abs(val_677)
        expand_5 = opset18.Expand(unsqueeze_16, val_679)
        val_683 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_684 = opset18.Concat(val_6, [40], val_683, [128], axis=0)
        _unsafe_view_2 = opset18.Reshape(expand_5, val_684, allowzero=1)
        unsqueeze_17 = opset18.Unsqueeze(cat_12, [2])
        val_729 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_730 = opset18.Concat(val_6, [10], [4], val_729, [128], axis=0)
        val_732 = opset18.Abs(val_730)
        expand_6 = opset18.Expand(unsqueeze_17, val_732)
        val_736 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_737 = opset18.Concat(val_6, [40], val_736, [128], axis=0)
        _unsafe_view_3 = opset18.Reshape(expand_6, val_737, allowzero=1)
        transpose_9 = opset18.Transpose(_unsafe_view_2, perm=[0, 1, 3, 2])
        matmul_3 = opset18.MatMul(add_720, transpose_9)
        mul_814 = opset18.Mul(matmul_3, 0.088388346)
        val_757 = opset18.Constant(value_ints=[0])
        val_759 = opset18.Constant(value_ints=[-1])
        val_760 = opset18.Reshape(add_4, val_759, allowzero=0)
        val_764 = opset18.Constant(value_ints=[1])
        slice_75 = opset18.Slice(slice_scatter_2, val_757, val_760, [3], val_764)
        add_907 = opset18.Add(mul_814, slice_75)
        val_765 = opset18.Softmax(add_907, axis=-1)
        matmul_4 = opset18.MatMul(val_765, _unsafe_view_3)
        transpose_10 = opset18.Transpose(matmul_4, perm=[0, 2, 1, 3])
        val_770 = opset18.Concat(val_6, val_1, [-1], axis=0)
        view_8 = opset18.Reshape(transpose_10, val_770, allowzero=1)
        val_772 = opset18.Transpose(model_layers_1_self_attn_o_proj_weight, perm=[1, 0])
        linear_5 = opset18.MatMul(view_8, val_772)
        add_950 = opset18.Add(add_592, linear_5)
        pow_4 = opset18.Pow(add_950, 2.0)
        mean_3 = opset18.ReduceMean(pow_4, [-1], noop_with_empty_axes=0, keepdims=1)
        add_963 = opset18.Add(mean_3, 1e-05)
        val_776 = opset18.Sqrt(add_963)
        rsqrt_3 = opset18.Reciprocal(val_776)
        mul_887 = opset18.Mul(add_950, rsqrt_3)
        mul_891 = opset18.Mul(model_layers_1_post_attention_layernorm_weight, mul_887)
        val_777 = opset18.Transpose(model_layers_1_mlp_gate_up_proj_weight, perm=[1, 0])
        linear_6 = opset18.MatMul(mul_891, val_777)
        split_1_split_0, split_1_split_1 = opset18.Split(linear_6, axis=2, num_outputs=2)
        val_778 = opset18.Sigmoid(split_1_split_0)
        silu_1 = opset18.Mul(split_1_split_0, val_778)
        mul_907 = opset18.Mul(split_1_split_1, silu_1)
        val_779 = opset18.Transpose(model_layers_1_mlp_down_proj_weight, perm=[1, 0])
        linear_7 = opset18.MatMul(mul_907, val_779)
        add_1008 = opset18.Add(add_950, linear_7)
        pow_5 = opset18.Pow(add_1008, 2.0)
        mean_4 = opset18.ReduceMean(pow_5, [-1], noop_with_empty_axes=0, keepdims=1)
        add_1021 = opset18.Add(mean_4, 1e-05)
        val_783 = opset18.Sqrt(add_1021)
        rsqrt_4 = opset18.Reciprocal(val_783)
        mul_929 = opset18.Mul(add_1008, rsqrt_4)
        mul_933 = opset18.Mul(model_norm_weight, mul_929)
        val_804 = opset18.Transpose(lm_head_weight, perm=[1, 0])
        linear_8 = opset18.MatMul(mul_933, val_804)
        return linear_8, cat_5, cat_11, cat_6, cat_12

    model = main_graph.to_model_proto(value_infos=value_infos)
    return model


def make_model_with_random_weights():
    model_embed_tokens_weight = numpy.random.rand(100352, 5120).astype(numpy.float32)
    model_layers_0_self_attn_o_proj_weight = numpy.random.rand(5120, 5120).astype(
        numpy.float32
    )
    model_layers_0_self_attn_qkv_proj_weight = numpy.random.rand(7680, 5120).astype(
        numpy.float32
    )
    model_layers_0_mlp_gate_up_proj_weight = numpy.random.rand(35840, 5120).astype(
        numpy.float32
    )
    model_layers_0_mlp_down_proj_weight = numpy.random.rand(5120, 17920).astype(numpy.float32)
    model_layers_0_input_layernorm_weight = numpy.random.rand(5120).astype(numpy.float32)
    model_layers_0_post_attention_layernorm_weight = numpy.random.rand(5120).astype(
        numpy.float32
    )
    model_layers_1_self_attn_o_proj_weight = numpy.random.rand(5120, 5120).astype(
        numpy.float32
    )
    model_layers_1_self_attn_qkv_proj_weight = numpy.random.rand(7680, 5120).astype(
        numpy.float32
    )
    model_layers_1_mlp_gate_up_proj_weight = numpy.random.rand(35840, 5120).astype(
        numpy.float32
    )
    model_layers_1_mlp_down_proj_weight = numpy.random.rand(5120, 17920).astype(numpy.float32)
    model_layers_1_input_layernorm_weight = numpy.random.rand(5120).astype(numpy.float32)
    model_layers_1_post_attention_layernorm_weight = numpy.random.rand(5120).astype(
        numpy.float32
    )
    model_norm_weight = numpy.random.rand(5120).astype(numpy.float32)
    lm_head_weight = numpy.random.rand(100352, 5120).astype(numpy.float32)
    expand_2 = numpy.random.rand(1, 64, 1).astype(numpy.float32)
    model = make_model(
        model_embed_tokens_weight,
        model_layers_0_self_attn_o_proj_weight,
        model_layers_0_self_attn_qkv_proj_weight,
        model_layers_0_mlp_gate_up_proj_weight,
        model_layers_0_mlp_down_proj_weight,
        model_layers_0_input_layernorm_weight,
        model_layers_0_post_attention_layernorm_weight,
        model_layers_1_self_attn_o_proj_weight,
        model_layers_1_self_attn_qkv_proj_weight,
        model_layers_1_mlp_gate_up_proj_weight,
        model_layers_1_mlp_down_proj_weight,
        model_layers_1_input_layernorm_weight,
        model_layers_1_post_attention_layernorm_weight,
        model_norm_weight,
        lm_head_weight,
        expand_2,
    )
    return model


class _Phi4LMTest:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model


def phi4lm_test():
    return _Phi4LMTest()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Stable APIs for PyTorch 2.6."""

from __future__ import annotations

__all__ = [
    "check_model",
    "convert_version",
    "get_torchlib_ops",
    "optimize",
    "save_model_with_external_data",
    "torchlib_opset",
]
import logging
from typing import TYPE_CHECKING

from onnxscript import ir, optimizer, version_converter
from onnxscript._framework_apis.torch_2_5 import (
    check_model,
    get_torchlib_ops,
    save_model_with_external_data,
)

if TYPE_CHECKING:
    from onnxscript.onnx_opset._impl.opset18 import Opset18


logger = logging.getLogger(__name__)


def optimize(model: ir.Model) -> ir.Model:
    """Optimize the model."""
    optimizer.optimize_ir(model)
    return model


def convert_version(model: ir.Model, target_version: int) -> ir.Model:
    """Convert the model to the specified ONNX opset version."""
    if target_version < 18:
        logger.warning("Conversion to opset < 18 is not supported.")
        return model
    version_converter.convert_version(model, target_version, fallback=True)
    return model


def torchlib_opset() -> Opset18:
    """Return the default opset for torchlib."""
    import onnxscript  # pylint: disable=import-outside-toplevel

    return onnxscript.opset18  # type: ignore


def torchlib_opset_version() -> int:
    """Return the default opset version for torchlib."""

    return torchlib_opset().version



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Stable APIs for PyTorch 2.5."""

from __future__ import annotations

__all__ = [
    "check_model",
    "convert_version",
    "get_torchlib_ops",
    "optimize",
    "save_model_with_external_data",
]

import dataclasses
import importlib.util
import os
import pathlib
from typing import Callable

from onnxscript import ir, optimizer, version_converter
from onnxscript.function_libs.torch_lib import registration


@dataclasses.dataclass(frozen=True)
class _OnnxFunctionMeta:
    """A wrapper of onnx-script function with additional metadata.

    qualified_name: The qualified name of the aten operator.
    function: The onnx-script function.
    domain: The domain of the function.
    name: The name of the function.
    is_complex: Whether the function is a complex function.
    """

    qualified_name: str
    function: Callable
    domain: str
    name: str
    is_complex: bool = False


def optimize(model: ir.Model) -> ir.Model:
    """Optimize the model."""
    # Internal flag. Will go away.
    enabled = os.getenv("TORCH_ONNX_ENABLE_OPTIMIZATION") == "1"
    if enabled:
        optimizer.optimize_ir(model)
    return model


def convert_version(model: ir.Model, target_version: int) -> ir.Model:
    """Convert the model to the specified ONNX opset version."""
    # Internal flag. Will go away.
    enabled = os.getenv("TORCH_ONNX_ENABLE_VERSION_CONVERSION") == "1"
    if enabled:
        version_converter.convert_version(model, target_version)
    return model


def check_model(model: ir.Model) -> None:
    """Check the model."""

    del model  # Unused yet


def save_model_with_external_data(
    model: ir.Model, model_path: str | os.PathLike, verbose: bool = False
) -> None:
    """Save the model with external data. The model is unchanged after saving."""

    # TODO(#1835): Decide if we want to externalize large attributes as well
    uninitialized_values = [
        value.name for value in model.graph.initializers.values() if value.const_value is None
    ]
    if uninitialized_values:
        raise ValueError(
            f"The model contains uninitialized initializer values ({uninitialized_values}). "
            "Please make sure all initializer values are initialized."
        )
    destination_path = pathlib.Path(model_path)
    data_path = f"{destination_path.name}.data"

    # Show a progress bar if verbose is True and tqdm is installed
    use_tqdm = verbose and importlib.util.find_spec("tqdm") is not None

    if use_tqdm:
        import tqdm  # pylint: disable=import-outside-toplevel

        with tqdm.tqdm() as pbar:
            total_set = False

            def callback(
                tensor: ir.TensorProtocol, metadata: ir.external_data.CallbackInfo
            ) -> None:
                nonlocal total_set
                if not total_set:
                    pbar.total = metadata.total
                    total_set = True

                pbar.update()
                pbar.set_description(
                    f"Saving {tensor.name} ({tensor.dtype.short_name()}, {tensor.shape}) at offset {metadata.offset}"
                )

            ir.save(model, model_path, external_data=data_path, callback=callback)
    else:
        ir.save(model, model_path, external_data=data_path)


def get_torchlib_ops() -> list[_OnnxFunctionMeta]:
    # Trigger op registration
    from onnxscript.function_libs.torch_lib import (  # pylint: disable=import-outside-toplevel
        ops,
    )

    del ops  # Unused

    torchlib_registry = registration.default_registry
    function_metas = []

    for qualified_name, aten_overloads_func in torchlib_registry.items():
        if qualified_name.startswith("internal::"):
            # Skip the custom defined internal functions
            continue

        for overload_func in aten_overloads_func.overloads:
            function_meta = _OnnxFunctionMeta(
                qualified_name=qualified_name,
                function=overload_func,
                domain=overload_func.function_ir.domain,
                name=overload_func.name,
                is_complex=False,
            )
            function_metas.append(function_meta)
        for complex_func in aten_overloads_func.complex:
            function_meta = _OnnxFunctionMeta(
                qualified_name=qualified_name,
                function=complex_func,
                domain=complex_func.function_ir.domain,
                name=complex_func.name,
                is_complex=True,
            )
            function_metas.append(function_meta)

    return function_metas



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
__all__ = [
    "core",
    "fft",
    "linalg",
    "nested",
    "nn",
    "prims",
    "quantized_decomposed",
    "sparse",
    "special",
    "vision",
]

from . import (
    core,
    fft,
    linalg,
    nested,
    nn,
    prims,
    quantized_decomposed,
    sparse,
    special,
    vision,
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnxscript.optimizer
from onnxscript.rewriter.models._bart_encoder import bart_encoder_test
from onnxscript.rewriter.models._smollm_1 import smollm_test_1
from onnxscript.rewriter.models._whisper_decoder import whisper_decoder_test
from onnxscript.rewriter.models._whisper_encoder import whisper_encoder_test
from onnxscript.rewriter.ort_fusions._test_utils import assert_allclose, ort_run
from onnxscript.rewriter.ort_fusions.rms_normalization import fuse_rms_normalization
from onnxscript.rewriter.ort_fusions.skip_normalization import (
    fuse_skip_layer_normalization,
    fuse_skip_rms_normalization,
)


class TestSkipNormalization(unittest.TestCase):
    def test_smollm(self):
        smollm_test = smollm_test_1()
        model = smollm_test.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        inputs = smollm_test.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)
        fuse_rms_normalization(model)
        fuse_skip_rms_normalization(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("SkipSimplifiedLayerNormalization", op_types)
        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)

    @unittest.skip("fixme: accuracy is not high")
    def test_whisper_encoder(self):
        whisper_encoder = whisper_encoder_test()
        model = whisper_encoder.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        inputs = whisper_encoder.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)

        fuse_skip_layer_normalization(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("SkipLayerNormalization", op_types)

        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)

    def test_whisper_decoder(self):
        whisper_decoder = whisper_decoder_test()
        model = whisper_decoder.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        inputs = whisper_decoder.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)

        fuse_skip_layer_normalization(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("SkipLayerNormalization", op_types)

        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)

    def test_bart_encoder(self):
        bart_encoder = bart_encoder_test()
        model = bart_encoder.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        inputs = bart_encoder.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)

        fuse_skip_layer_normalization(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("SkipLayerNormalization", op_types)
        self.assertEqual(op_types.count("SkipLayerNormalization"), 5)
        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import onnx.parser
import parameterized

from onnxscript import ir
from onnxscript.rewriter.ort_fusions import softmax


class SoftmaxUpcastRemovalTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ("Softmax<axis=-1>",),
            ("Softmax",),
        ]
    )
    def test_softmax_upcast_to_fp32_is_removed_when_input_and_final_output_is_fp16(
        self, softmax_op_str
    ):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float16[N] x) => (float16[N] z)
            {{
                x_fp32 = Cast<to=1>(x)
                z_fp32 = {softmax_op_str}(x_fp32)
                z = Cast<to=10>(z_fp32)
            }}
            """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = softmax.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertNotIn("Cast", {node.op_type for node in model.graph})

    @parameterized.parameterized.expand(
        [
            ("Softmax<axis=-1>",),
            ("Softmax",),
        ]
    )
    def test_softmax_upcast_to_fp32_is_not_removed_when_input_is_not_fp16(
        self, softmax_op_str
    ):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (int32[N] x) => (float16[N] z)
            {{
                x_fp32 = Cast<to=1>(x)
                z_fp32 = {softmax_op_str}(x_fp32)
                z = Cast<to=10>(z_fp32)
            }}
            """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = softmax.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(
            len([node.op_type for node in model.graph if node.op_type == "Cast"]), 2
        )

    @parameterized.parameterized.expand(
        [
            ("Softmax<axis=-1>",),
            ("Softmax",),
        ]
    )
    def test_softmax_upcast_to_fp32_is_not_removed_when_final_output_is_not_fp16(
        self, softmax_op_str
    ):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float16[N] x) => (double[N] z)
            {{
                x_fp32 = Cast<to=1>(x)
                z_fp32 = {softmax_op_str}(x_fp32)
                z = Cast<to=11>(z_fp32)
            }}
            """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = softmax.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(
            len([node.op_type for node in model.graph if node.op_type == "Cast"]), 2
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest
from typing import Any

import numpy as np
import onnx
import onnx.reference
import parameterized

import onnxscript
import onnxscript.onnx_types as ot
from onnxscript import ir
from onnxscript.onnx_opset import opset18
from onnxscript.rewriter import MatchingTracer, testing
from onnxscript.rewriter import pattern as orp
from onnxscript.rewriter.rules.common import _basic_rules

FLOAT = onnx.TensorProto.FLOAT


@onnxscript.script()
def cast_identity_model(x: ot.FLOAT["a", "b", "c"]) -> ot.FLOAT["a", "b", "c"]:  # noqa: F821, UP037
    y = opset18.Cast(x, to=onnx.TensorProto.FLOAT)
    return y


def _make_model(*args, **kwargs) -> ir.Model:
    return ir.serde.deserialize_model(onnx.helper.make_model(*args, **kwargs))


def clone_model(model: ir.Model) -> ir.Model:
    return ir.from_proto(ir.to_proto(model))


class BasicRulesTest(unittest.TestCase):
    def _get_random_inputs(self, model: onnx.ModelProto) -> dict[str, Any]:
        feeds: dict[str, Any] = {}
        for i in model.graph.input:
            ish = tuple(i.type.tensor_type.shape.dim)
            # Creates an input tensor with a dimension defined by the onnx model
            # or equals to i + 2 with i being the dimension index.
            # The tensor is kept small to make the test fast.
            shape = tuple(
                (d.dim_value if d.dim_value > 0 else i + 2) for i, d in enumerate(ish)
            )
            if i.type.tensor_type.elem_type == onnx.TensorProto.FLOAT:
                feeds[i.name] = np.random.randn(*shape).astype(np.float32)
            else:
                raise AssertionError(f"Not implemented for input {i}")
        return feeds

    def _check_model(
        self,
        model: onnx.ModelProto,
        optimized_model: onnx.ModelProto,
        feeds: dict[str, Any] | None = None,
        atol: float = 0.0,
        rtol: float = 1e-7,
    ):
        if not feeds:
            feeds = self._get_random_inputs(model)
        ref = onnx.reference.ReferenceEvaluator(model)
        opt = onnx.reference.ReferenceEvaluator(optimized_model)
        expected = ref.run(None, feeds)
        got = opt.run(None, feeds)
        self.assertEqual(len(expected), len(got))
        for a, b in zip(expected, got):
            np.testing.assert_allclose(a, b, atol=atol, rtol=rtol)

    @parameterized.parameterized.expand(
        [
            (
                "no_op_transpose",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Transpose", ["X"], ["Y"], perm=[0, 1, 2]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [None, None, None])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [None, None, None])],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
            (
                "canceled_out_transposes",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Transpose", ["X"], ["xt"], perm=[1, 0]),
                            onnx.helper.make_node("Transpose", ["xt"], ["Y"], perm=[1, 0]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [None, None])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [None, None])],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
        ]
    )
    def test_basic_optimization_rules_identity(self, _: str, model: ir.Model):
        rule_set = _basic_rules.basic_optimization_rules()
        model_proto = ir.serde.serialize_model(model)
        rule_set.apply_to_model(model)
        rewritten_model = ir.serde.serialize_model(model)

        self.assertEqual(["Identity"], [n.op_type for n in model.graph])
        self._check_model(model_proto, rewritten_model)

    @parameterized.parameterized.expand(
        [
            (
                "consecutive_transposes",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Transpose", ["X"], ["xt"], perm=[1, 2, 0]),
                            onnx.helper.make_node("Transpose", ["xt"], ["Y"], perm=[1, 2, 0]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [None, None, None])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [None, None, None])],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
        ]
    )
    def test_basic_optimization_rules_transpose_transpose(self, _: str, model: ir.Model):
        rule_set = _basic_rules.basic_optimization_rules()
        model_proto = ir.serde.serialize_model(model)
        rule_set.apply_to_model(model)
        rewritten_model = ir.serde.serialize_model(model)
        self.assertEqual(["Transpose"], [n.op_type for n in model.graph])
        self._check_model(model_proto, rewritten_model)

    def _double_cast_model(self, ostype1, ostype2, ostype3):
        dtype2 = ostype2.dtype
        dtype3 = ostype3.dtype

        @onnxscript.script()
        def cast_cast_model(x):
            intermediate = opset18.Cast(x, to=dtype2)
            y = opset18.Cast(intermediate, to=dtype3)
            return y

        return cast_cast_model.to_model_proto(
            input_types=[ostype1[10]], output_types=[ostype3[10]]
        )

    @parameterized.parameterized.expand(
        [
            ("float16_float_float16", ot.FLOAT16, ot.FLOAT, ot.FLOAT16),
        ]
    )
    def test_cast_cast_rule(self, _: str, type1, type2, type3):
        rule = _basic_rules.cast_cast_rule
        model_proto = self._double_cast_model(type1, type2, type3)
        model = ir.serde.deserialize_model(model_proto)
        rule.apply_to_model(model)
        _rewritten_model = ir.serde.serialize_model(model)

        self.assertEqual(["Cast"], [n.op_type for n in model.graph])
        # TODO: (random) fp16 inputs
        # self._check_model(model_proto, rewritten_model, atol=1e-2)

    @parameterized.parameterized.expand(
        [
            (
                "cast_identity",
                ir.serde.deserialize_model(cast_identity_model.to_model_proto()),
            ),
        ]
    )
    def test_cast_identity_rule(self, _: str, model: ir.Model):
        rule_set = _basic_rules.basic_optimization_rules()
        model_proto = ir.serde.serialize_model(model)
        rule_set.apply_to_model(model)
        rewritten_model = ir.serde.serialize_model(model)

        self.assertEqual(["Identity"], [n.op_type for n in model.graph])
        self._check_model(model_proto, rewritten_model)

    @parameterized.parameterized.expand(
        [
            (
                "normal_case",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Expand", ["X", "shape"], ["Y"]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [3, 4, 5])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [3, 4, 5])],
                        [
                            onnx.numpy_helper.from_array(
                                np.array([3, 4, 5], dtype=np.int64), name="shape"
                            )
                        ],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
                ("Identity",),
            ),
            (
                "input_no_shape",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Identity", ["X"], ["Y"]),
                            onnx.helper.make_node("Expand", ["Y", "shape"], ["Z"]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [3, 4, 5])],
                        [onnx.helper.make_tensor_value_info("Z", FLOAT, [3, 4, 5])],
                        [
                            onnx.numpy_helper.from_array(
                                np.array([3, 4, 5], dtype=np.int64), name="shape"
                            )
                        ],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
                ("Identity", "Expand"),
            ),
        ]
    )
    def test_expand_identity_rule(
        self, _: str, model: ir.Model, expected_nodes: tuple[str, ...]
    ):
        rule_set = _basic_rules.basic_optimization_rules()
        model_proto = ir.serde.serialize_model(model)
        rule_set.apply_to_model(model)
        rewritten_model = ir.serde.serialize_model(model)

        self.assertEqual(tuple(n.op_type for n in model.graph), expected_nodes)
        self._check_model(model_proto, rewritten_model)

    @parameterized.parameterized.expand(
        [
            (
                "double_unsqueezes_1",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Unsqueeze", ["X", "axes1"], ["Xu"]),
                            onnx.helper.make_node("Unsqueeze", ["Xu", "axes2"], ["Y"]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [3])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [1, 3, 1])],
                        [
                            onnx.numpy_helper.from_array(
                                np.array([1], dtype=np.int64), name="axes1"
                            ),
                            onnx.numpy_helper.from_array(
                                np.array([0], dtype=np.int64), name="axes2"
                            ),
                        ],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
            (
                "double_unsqueezes_2",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Unsqueeze", ["X", "axes1"], ["Xu"]),
                            onnx.helper.make_node("Unsqueeze", ["Xu", "axes2"], ["Y"]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [3])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [1, 3, 1])],
                        [
                            onnx.numpy_helper.from_array(
                                np.array([0], dtype=np.int64), name="axes1"
                            ),
                            onnx.numpy_helper.from_array(
                                np.array([1], dtype=np.int64), name="axes2"
                            ),
                        ],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
            (
                "double_unsqueezes_3",
                _make_model(
                    onnx.helper.make_graph(
                        [
                            onnx.helper.make_node("Unsqueeze", ["X", "axes1"], ["Xu"]),
                            onnx.helper.make_node("Unsqueeze", ["Xu", "axes2"], ["Y"]),
                        ],
                        "name",
                        [onnx.helper.make_tensor_value_info("X", FLOAT, [3])],
                        [onnx.helper.make_tensor_value_info("Y", FLOAT, [1, 3, 1])],
                        [
                            onnx.numpy_helper.from_array(
                                np.array(0, dtype=np.int64), name="axes1"
                            ),
                            onnx.numpy_helper.from_array(
                                np.array(1, dtype=np.int64), name="axes2"
                            ),
                        ],
                    ),
                    opset_imports=[onnx.helper.make_opsetid("", 18)],
                ),
            ),
        ]
    )
    def test_unsqueeze_unsqueeze_rule(self, _: str, model: ir.Model):
        rule_set = _basic_rules.basic_optimization_rules()
        model_proto = ir.serde.serialize_model(model)
        rule_set.apply_to_model(model)
        rewritten_model = ir.serde.serialize_model(model)

        self.assertEqual(["Constant", "Unsqueeze"], [n.op_type for n in model.graph])
        self._check_model(model_proto, rewritten_model)

    @classmethod
    def _slices_split_models(cls):
        models = [
            _make_model(
                onnx.helper.make_graph(
                    [
                        onnx.helper.make_node(
                            "Slice", ["X", "zero", "half", "axis"], ["spl1"]
                        ),
                        onnx.helper.make_node(
                            "Slice", ["X", "half", "last", "axis"], ["spl2"]
                        ),
                    ],
                    "name",
                    [onnx.helper.make_tensor_value_info("X", FLOAT, [3, 4, 6])],
                    [
                        onnx.helper.make_tensor_value_info("spl1", FLOAT, [3, 4, 3]),
                        onnx.helper.make_tensor_value_info("spl2", FLOAT, [3, 4, 3]),
                    ],
                    [
                        onnx.numpy_helper.from_array(
                            np.array([0], dtype=np.int64), name="zero"
                        ),
                        onnx.numpy_helper.from_array(
                            np.array([3], dtype=np.int64), name="half"
                        ),
                        onnx.numpy_helper.from_array(
                            np.array([6], dtype=np.int64), name="last"
                        ),
                        onnx.numpy_helper.from_array(
                            np.array([2], dtype=np.int64), name="axis"
                        ),
                    ],
                ),
                opset_imports=[onnx.helper.make_opsetid("", 18)],
            ),
        ]
        return models

    @unittest.skipIf(True, reason="see https://github.com/microsoft/onnxscript/issues/1642")
    def test_slices_split_rule(self):
        for model_proto in self._slices_split_models():
            ir_model = ir.serde.deserialize_model(model_proto)
            rule_set = _basic_rules.basic_optimization_rules()
            rule_set.apply_to_model(ir_model)
            rewritten_model = ir.serde.serialize_model(ir_model)

            self.assertEqual(["Split"], [n.op_type for n in rewritten_model.graph.node])
            self._check_model(model_proto, rewritten_model)

    def test_squeeze_reshape_1d_rule(self):
        rule = _basic_rules.squeeze_reshape_1d_rule

        def check(model_script, expected_count) -> None:
            model_proto = model_script.to_model_proto()
            ir_model = ir.serde.deserialize_model(model_proto)
            count = rule.apply_to_model(ir_model)
            self.assertEqual(count, expected_count)
            if count > 0:
                self.assertEqual([x.op_type for x in ir_model.graph], ["Identity"])
            rewritten_proto = ir.serde.serialize_model(ir_model)
            self._check_model(model_proto, rewritten_proto)

        op = onnxscript.opset17

        # input of shape [12]
        @onnxscript.script()
        def model1(X: ot.FLOAT[12]):
            return op.Reshape(op.Squeeze(X), [-1])

        check(model1, 1)

        # input of shape [1]
        @onnxscript.script()
        def model2(X: ot.FLOAT[1]):
            return op.Reshape(op.Squeeze(X), [-1])

        check(model2, 1)

        # input of shape [1, 1]
        # This should NOT be optimized to Identity
        @onnxscript.script()
        def model3(X: ot.FLOAT[1, 1]):
            return op.Reshape(op.Squeeze(X), [-1])

        check(model3, 0)


class ReshapeReshapeTest(unittest.TestCase):
    @staticmethod
    def create_model(
        input_shape, shape1, shape2, allowzero1=0, allowzero2=0, infer_shape=False
    ):
        def _convert_shape(shape, name):
            if isinstance(shape, np.ndarray):
                shape = tape.initializer(ir.Tensor(shape, name=name))
            elif isinstance(shape, (list, tuple)):
                shape = ir.val(name, ir.DataType.INT64, ir.Shape(shape))
                tape.graph_like.inputs.append(shape)
            else:
                raise TypeError(f"Unsupported type {type(shape)} for shape.")
            return shape

        x = ir.val("X", ir.DataType.FLOAT, ir.Shape(input_shape))
        y = ir.val("Y", ir.DataType.FLOAT)
        tape = ir.tape.Tape(ir.Graph([x], [y], nodes=[], opset_imports={"": 20}))

        # Build the graph.
        reshape = tape.op(
            "Reshape",
            inputs=[x, _convert_shape(shape1, "shape_")],
            attributes={"allowzero": allowzero1},
        )
        tape.op(
            "Reshape",
            inputs=[reshape, _convert_shape(shape2, "shape")],
            attributes={"allowzero": allowzero2},
            output=y,
        )
        model = ir.Model(tape.graph_like, ir_version=10)

        # Infer shapes.
        if infer_shape:
            model = ir.passes.common.ShapeInferencePass()(model).model
        return model

    @parameterized.parameterized.expand(
        [
            ((3, 4, 5), [4, 5, 3], [5, 4, 3]),
            ((3, 4, 5), [4, 5, 3], [5, 4, 3]),
            ((3, 4, 8), [2, 0, 3, -1], [0, 3, 2, 8]),
            ((3, 4, 8), [3, 4, -1], [-1, 12], 1),
            ((3, 4, 2), [0, 4, -1], [12, -1], 0, 1),
            ((3, 0, 8), [4, 2, 0, 0], [3, 0], 1, 1),
        ]
    )
    def test_reshape_reshape_rule(
        self, input_shape, shape1, shape2, allowzero1=0, allowzero2=0
    ):
        model = self.create_model(
            input_shape,
            np.array(shape1, dtype="int64"),
            np.array(shape2, dtype="int64"),
            allowzero1=allowzero1,
            allowzero2=allowzero2,
        )
        updated_model = clone_model(model)

        # check rewrite approach.
        count = _basic_rules.reshape_reshape_rule.apply_to_model(updated_model)
        self.assertEqual(count, 1)
        self.assertEqual(["Reshape"], [n.op_type for n in updated_model.graph])

        # Check inference.
        inputs = np.random.default_rng(10).random(input_shape, dtype="float32")
        testing.assert_numerically_equal(model, updated_model, (inputs,), atol=0, rtol=0)

    @parameterized.parameterized.expand([([3, 2, 3, 3, 3], 1), ([0, -1, 3, 2], 0)])
    def test_reshape_dynamic_reshape_rule(self, shape1, allowzero1=0):
        input_shape = (3, 6, 9)
        shape1 = np.array(shape1, dtype="int64")
        # Build the model with unknown shape1.
        model = self.create_model(
            input_shape,
            (shape1.size,),
            np.array((1, 6, 27), dtype="int64"),
            allowzero1=allowzero1,
        )
        updated_model = clone_model(model)

        # check rewrite approach.
        count = _basic_rules.reshape_reshape_rule.apply_to_model(updated_model)
        self.assertEqual(count, 1)
        self.assertEqual(["Reshape"], [n.op_type for n in updated_model.graph])

        # Check inference.
        feeds = {
            "X": np.random.default_rng(2).random(input_shape, dtype="float32"),
            "shape_": shape1,
        }
        testing.assert_numerically_equal(model, updated_model, feeds, atol=0, rtol=0)

    @parameterized.parameterized.expand(
        [((3, 6, 9), [0, 3, 2, -1]), ((0, 6, 2), [0, 0, 3], 1)]
    )
    def test_reshape_reshape_dynamic_rule(self, input_shape, shape2, allowzero2=0):
        # Note that shape inference is required for this test to be valid.
        shape2 = np.array(shape2, dtype="int64")
        model = self.create_model(
            input_shape,
            np.array((3, 2, -1), dtype="int64"),
            shape2,
            allowzero2=allowzero2,
            infer_shape=True,
        )
        updated_model = clone_model(model)

        # check rewrite approach.
        count = _basic_rules.reshape_reshape_rule.apply_to_model(updated_model)
        self.assertEqual(count, 1)
        self.assertEqual(["Reshape"], [n.op_type for n in updated_model.graph])

        # Check inference.
        inputs = np.random.default_rng(7).random(input_shape, dtype="float32")
        testing.assert_numerically_equal(model, updated_model, (inputs,), atol=0, rtol=0)

    @parameterized.parameterized.expand(
        [
            ((3,), "is not a constant"),
            (np.array([0, -1], dtype="int64"), "both 0 and -1 dimensions"),
            (np.array([0, 0, 3], dtype="int64"), "more than one 0 dimension"),
        ]
    )
    def test_unsupported_reshape_reshape(self, shape2, error_msg):
        model = self.create_model((1, 2, 3), np.array([1, 6], dtype="int64"), shape2)

        # Check rewrite approach.
        tracer = MatchingTracer()
        count = _basic_rules.reshape_reshape_rule.apply_to_model(model, tracer=tracer)
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[_basic_rules.reshape_reshape_rule][0]
        self.assertEqual(tracer_match.status.value, orp.MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, error_msg)


class Flatten2ReshapeTest(unittest.TestCase):
    @staticmethod
    def create_model(input_shape, axis=1):
        x = ir.val("X", ir.DataType.FLOAT, ir.Shape(input_shape))
        y = ir.val("Y", ir.DataType.FLOAT)
        tape = ir.tape.Tape(ir.Graph([x], [y], nodes=[], opset_imports={"": 20}))

        # Build the graph.
        tape.op("Flatten", inputs=[x], attributes={"axis": axis}, output=y)
        model = ir.Model(tape.graph_like, ir_version=10)
        return model

    @parameterized.parameterized.expand(list(range(-5, 6)))
    def test_flatten_to_reshape_rule(self, axis):
        input_shape = (1, 4, 8, 7, 5)
        model = self.create_model(input_shape=input_shape, axis=axis)
        updated_model = clone_model(model)

        # check rewrite approach.
        count = _basic_rules.flatten_to_reshape_rule.apply_to_model(updated_model)
        self.assertEqual(count, 1)
        self.assertEqual(["Reshape"], [n.op_type for n in updated_model.graph])

        # Check inference.
        inputs = np.random.default_rng(13).random(input_shape, dtype="float32")
        testing.assert_numerically_equal(model, updated_model, (inputs,), atol=0, rtol=0)

    @parameterized.parameterized.expand(list(range(-4, 5)))
    def test_flatten_to_reshape_dynamic_input(self, axis):
        model = self.create_model(input_shape=("N", "C1", "C2", "C3"), axis=axis)
        # Rule is supported in all cases if the output shape is known for non-special cases.
        input_shape = (1, 2, 3, 4)
        if axis not in {-3, 0, 1, 4}:
            out_shape = ir.Shape((np.prod(input_shape[:axis]), np.prod(input_shape[axis:])))
            model.graph.outputs[0].shape = out_shape
        updated_model = clone_model(model)

        # check rewrite approach.
        count = _basic_rules.flatten_to_reshape_rule.apply_to_model(updated_model)
        self.assertEqual(count, 1)
        self.assertEqual(["Reshape"], [n.op_type for n in updated_model.graph])

        # Check inference.
        inputs = np.random.default_rng(17).random(input_shape, dtype="float32")
        testing.assert_numerically_equal(model, updated_model, (inputs,), atol=0, rtol=0)

    def test_unsupported_flatten_to_reshape(self):
        model = self.create_model(input_shape=("N", "C1", "C2"), axis=2)

        # Check rewrite approach.
        tracer = MatchingTracer()
        count = _basic_rules.flatten_to_reshape_rule.apply_to_model(model, tracer=tracer)
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[_basic_rules.flatten_to_reshape_rule][0]
        self.assertEqual(tracer_match.status.value, orp.MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, "Impossible to compute new shape")


if __name__ == "__main__":
    unittest.main(verbosity=2)
